import datetime
import glob
from math import log, sqrt, isnan
from netCDF4 import Dataset
import numpy as np
import numpy.ma as ma
import os
from os import listdir
from os.path import isfile, join, abspath
import re
import threading
import time
import collections

from collections import OrderedDict

import wx
import wx.lib.newevent

import nt_globals as NtGlobals
from nt_output import ResultPrinter, TextPrinter as TP
from nt_typing import NeuralProcessing, Election

# Processing result codes
RES_OK = 0
RES_WARNING = 1
RES_ERROR = 2
RES_ABORTED = 3

# Events generated by NtWorkerThread
WorkStarted, EVT_WORK_STARTED = wx.lib.newevent.NewEvent()
WorkAborted, EVT_WORK_ABORTED = wx.lib.newevent.NewEvent()
WorkFinished, EVT_WORK_FINISHED = wx.lib.newevent.NewEvent()
DatasetStarted, EVT_DATASET_STARTED = wx.lib.newevent.NewEvent()
DatasetFinished, EVT_DATASET_FINISHED = wx.lib.newevent.NewEvent()
DatasetWarning, EVT_DATASET_WARNING = wx.lib.newevent.NewEvent()
DatasetError, EVT_DATASET_ERROR = wx.lib.newevent.NewEvent()
ExtParamsRead, EVT_EXT_PARAMS_READ = wx.lib.newevent.NewEvent()
LayersComputed, EVT_LAYERS_COMPUTED = wx.lib.newevent.NewEvent()
IntParamsComputed, EVT_INT_PARAMS_COMPUTED = wx.lib.newevent.NewEvent()
AvgIntParamsComputed, EVT_AVG_INT_PARAMS_COMPUTED = wx.lib.newevent.NewEvent()
AerosolsComputed, EVT_AEROSOLS_COMPUTED = wx.lib.newevent.NewEvent()

class NtWorkerThread(threading.Thread):
    '''
    Thread class used to do all the processing work
    for a given set of datasets.
    '''
    
    def __init__(self, *args, **kwargs):
        '''
        Construct a NtWorkerThread object.
        
        Keyword arguments:
        parent -- wx.Panel parent of the NtWorkerThread.
        datasets -- list of datasets' names to be processed.
        folder -- full path of the folder where the datasets can be found.
        '''
        
        acceptedArgs = [
            'parent', 'datasets', 'folder', 'rep_file', 'csv_file',
            'ann_folder', 'filter_window', 'min_layer_depth',
            'finesse', 'min_accepted_confidence', 'min_agreement_percentage'
        ]
        
        self._rep_file = None
        self._csv_file = None
        self._ann_folder = NtGlobals.ANN_FOLDER
        self._filter_window = NtGlobals.FILTER_WINDOW
        self._min_layer_depth = NtGlobals.MIN_LAYER_DEPTH
        self._finesse = NtGlobals.FINESSE
        self._min_accepted_confidence = NtGlobals.MIN_ACCEPTED_CONFIDENCE
        self._min_agreement_percentage = NtGlobals.MIN_AGREEMENT_PERCENTAGE
        
        for arg, value in kwargs.iteritems():
            # Parse arguments and raise TypeError if any
            # unknown parameters are found.
            if arg not in acceptedArgs:
                raise TypeError('Invalid argument in NtWorkerThread.__init__(): '+arg)
                
            # Read known parameters and set member variables.
            if arg == 'parent':
                self._parent = value
            elif arg == 'datasets':
                self._datasets = value
            elif arg == 'folder':
                self._folder = value
            elif arg == 'rep_file':
                self._rep_file = value
            elif arg == 'csv_file':
                self._csv_file = value
            elif arg == 'ann_folder':
                self._ann_folder = value
            elif arg == 'filter_window':
                self._filter_window = value
            elif arg == 'min_layer_depth':
                self._min_layer_depth = value
            elif arg == 'finesse':
                self._finesse = value
            elif arg == 'min_accepted_confidence':
                self._min_accepted_confidence = value
            elif arg == 'min_agreement_percentage':
                self._min_agreement_percentage = value
                
        # Set optional parameters to default values if not
        # otherwise specified.
        
        if self._rep_file == None:
            self._rep_file = "%s.txt" % self._folder.split('\\')[-1]
            
        if self._csv_file == None:
            self._csv_file = "%s.csv" % self._folder.split('\\')[-1]
            
        self._rep_file = abspath(self._rep_file)
        self._csv_file = abspath(self._csv_file)
            
        self._wantToStop = False
            
        threading.Thread.__init__(self)
        
    def run(self):
        '''
        Starts processing the entire list of datasets provided in the constructor.
        '''
        
        startmsg="\nStart run time: %s\n" % datetime.datetime.today().strftime("%Y-%m-%d %H:%M")
        startmsg=("-"*(len(startmsg)-2))+startmsg+("-"*(len(startmsg)-2))+"\n"
        startmsg += 'Application version: %s\n' % NtGlobals.APPLICATION_TITLE
    
        TP.Log(startmsg)
        TP.PrintLine(NtGlobals.APPLICATION_TITLE)
    
        self._printer=ResultPrinter(True, self._csv_file, self._rep_file)
        
        self._printer.header(startmsg)
        
        global_start_time = time.time()
        TP.PrintLine('Processing %d measurements from folder %s' % (len(self._datasets), self._folder))
    
        header="\nInput\n-----\n"
        header+="Data folder: %s\n" % self._folder
        header+="Measurements:\n"
        
        for dataset in self._datasets:
            header += "\t%s\n" % dataset["measurement"]
        
        TP.Log(header)
        TP.Log("\nOutput\n------\n")
        TP.Log("CSV file path: %s\nReport file path: %s\n" % (self._csv_file, self._rep_file))
        
        header+="\nRun parameters\n------------------\n"
        header+="FILTER_WINDOW: %d\nMIN_LAYER_DEPTH: %d\n" % (self._filter_window, self._min_layer_depth)
        header+="FINESSE: %d\nANN_FOLDER: %s\n" % (self._finesse, self._ann_folder)
        header+="MIN_ACCEPTED_CONFIDENCE: %f\nMIN_AGREEMENT_PERCENTAGE: %f" % (self._min_accepted_confidence, self._min_agreement_percentage)
        
        self._printer.header(header)
        
        # Notify the parent we started working.
        wx.PostEvent(
            self._parent,
            WorkStarted(
                thread=threading.current_thread()
            )
        )
        
        self._progress = 0
        
        # Parse the entire list of datasets.
        for dataset in self._datasets:
        
            if self.Stopped():
                self._printer.footer(message="Aborted after %s seconds." % (round(time.time()-global_start_time,2)))
                
                wx.PostEvent(
                    self._parent,
                    WorkAborted(
                        thread=threading.current_thread()
                    )
                )
                
                self.CleanUp()
                return
        
            # Notify parent we started working on a new dataset.
            wx.PostEvent(
                self._parent,
                DatasetStarted(
                    data={'dataset': dataset["measurement"]},
                    thread=threading.current_thread()
                )
            )
            
            self._progress += 1
            
            start_time = time.time()
            
            self._printer.init_measurement(measurement=dataset["measurement"])
            
            padded_name = (dataset["measurement"]+(" (%d/%d)" % (self._progress, len(self._datasets)))).center(36)
            padded_name="|"+padded_name+"|"
        
            TP.PrintLine("\n+------------------------------------+")
            TP.PrintLine(padded_name)
            TP.PrintLine("+------------------------------------+")
        
            TP.PrintLine("Elapsed time: %d seconds\n" % (time.time()-global_start_time))
            
            # Do the processing.
            resultCode, resultMessage = self.ProcessDataset( dataset )
            
            # Check out the result of the processing.
            if resultCode == RES_ERROR:
                # Tell the parent there was an error processing
                # this dataset.
                wx.PostEvent(
                    self._parent,
                    DatasetError(
                        data={'dataset': dataset["measurement"], 'error': resultMessage},
                        thread=threading.current_thread()
                    )
                )
            else:
                # Tell the parent everything went OK for this dataset.
                wx.PostEvent(
                    self._parent,
                    DatasetFinished(
                        data={'dataset': dataset["measurement"]},
                        thread=threading.current_thread()
                    )
                )
                
            self._printer.measurement_footer(
                measurement=dataset["measurement"],
                message="\nProcessing time: %s seconds." % (round(time.time()-start_time,2))
            )
            
            TP.PrintLine("Processing took %s seconds." % (round(time.time()-start_time,2)))
        
        # Notify the parent we have finished our work.     
        wx.PostEvent(
            self._parent,
            WorkFinished(
                thread=threading.current_thread()
            )
        )
                
        self._printer.footer(message="Whole processing time: %s seconds." % (round(time.time()-global_start_time,2)))

        self.CleanUp()
        
    def stop(self):
        self._wantToStop = True
        
    def Stopped(self):
        return self._wantToStop
        
    def CleanUp(self):
        self._printer.close()
        
    def ProcessDataset(self, dataset):
        '''
        Do all the processing for one dataset.
        
        Keyword arguments:
        dataset -- the full path of the dataset you want to process.
        '''
        
        warningCount = 0
        warningMsg = ""
        
        TP.PrintLine("Checking measurement...")
        
        try:
            measurement=LidarMeasurement(dataset)
        except ValueError as err:
            TP.PrintLine("Error creating LidarMeasurement object! %s\n" % str(err))
            TP.Log("%s: Error creating LidarMeasurement object! %s\n" % (dataset["measurement"], str(err)))
            self._printer.measurement_header(dataset["measurement"], "\nError creating LidarMeasurement object! %s" % str(err))
            
            return RES_ERROR, "Error creating LidarMeasurement object! %s\n" % str(err)
            
        TP.PrintLine("Reading files...")
        
        if self.Stopped():
            return RES_ABORTED, 'Processing aborted!'
        
        try:
            measurement.read_data()
        except Exception as e:
            print str(e)
        
            TP.PrintLine("Error: %s Skipping measurement..." % (e))
            TP.Log("%s: Error: %s Skipping measurement...\n" % (dataset["measurement"], e))
            self._printer.measurement_header(dataset["measurement"], "\nError: %s Skipping measurement..." % (e))
            
            return RES_ERROR, "Error reading NetCDF files!"
            
        if self.Stopped():
            return RES_ABORTED, 'Processing aborted!'
            
        extparams=measurement.get_extparams()
        altitudes=measurement.get_altitudes()
        
        wx.PostEvent(
            self._parent,
            ExtParamsRead(
                data={'dataset': dataset["measurement"], 'params': extparams, 'altitudes': altitudes, 'has_dep': measurement.has_depolarisation()},
                thread=threading.current_thread()
            )
        )
        
        TP.PrintLine("Computing layers...")
        
        # Accepted parameters for .get_layers():
        # min_height, window_size, order
        try:
            layers=measurement.get_layers(min_height=self._min_layer_depth, window_size=self._filter_window)
        except:
            layers=[]
        
        
        if len(layers) < 1:
            TP.PrintLine("No layer found! Skipping measurement...")
            self._printer.measurement_header(dataset["measurement"], "\nNo layer found! Skipping measurement...")
            
            return RES_ERROR, "No layer found!"
        
        # You can further process/filter the detected layers here:
        ilayers=pick_layers(data=extparams["b1064"]["data"], layers=alt2index(altitudes, layers))
        layers=index2alt(altitudes,ilayers)
        
        wx.PostEvent(
            self._parent,
            LayersComputed(
                data={'dataset': dataset["measurement"], 'layers': layers, 'ilayers': ilayers},
                thread=threading.current_thread()
            )
        )
        
        if self.Stopped():
            return RES_ABORTED, 'Processing aborted!'
        
        compute_layers=[]
        
        TP.PrintLine("Computing intensive parameters...")
        
        measurement.compute_intensive_parameters(layers=layers)
        intparams=measurement.get_intparams()
        
        wx.PostEvent(
            self._parent,
            IntParamsComputed(
                data={'dataset': dataset["measurement"], 'params': intparams, 'altitudes': altitudes},
                thread=threading.current_thread()
            )
        )
        
        if self.Stopped():
            return RES_ABORTED, 'Processing aborted!'
        
        snr_layers=measurement.get_snr_layers(layers=layers, min_layer_height=self._min_layer_depth)
        
        if self.Stopped():
            return RES_ABORTED, 'Processing aborted!'
        
        avg_intparams=measurement.average_intparams(averaging_layers=snr_layers, layers=layers)
        
        if self.Stopped():
            return RES_ABORTED, 'Processing aborted!'
        
        self._printer.create_layers(measurement=dataset["measurement"], number_of_layers=len(avg_intparams))
        
        wx.PostEvent(
            self._parent,
            AvgIntParamsComputed(
                data={'dataset': dataset["measurement"], 'data': avg_intparams},
                thread=threading.current_thread()
            )
        )
        
        for i in range(0, len(avg_intparams)):
        
            if self.Stopped():
                return RES_ABORTED, 'Processing aborted!'
                
            self._printer.update_value(layer_number=i, measurement=dataset["measurement"], column="Retrieval Bottom", value=snr_layers[i][0])
            self._printer.update_value(layer_number=i, measurement=dataset["measurement"], column="Retrieval Top", value=snr_layers[i][1])
            
            layer = avg_intparams[i]
            param_list=[]
            
            translation_table={
                'ae_355_532': 'AE355_532',
                'ci_355_532': 'CI355_532',
                'ci_532_1064': 'CI532_1064',
                'cr_355_532': 'CR355_532',
                'cr_532_1064': 'CR532_1064',
                'lr_355': 'LR355',
                'lr_532': 'LR532',
            }
            
            if measurement.has_depolarisation():
                translation_table['dep'] = 'DEP532'
                
            for p1, p2 in translation_table.items():
                decimals=2
                
                if "LR" in p2:
                    msg="%.0f" % round(layer[p1]["average"])
                else:
                    msg="%.2f" % layer[p1]["average"]

                self._printer.update_value(layer_number=i, measurement=dataset["measurement"], column=p2, value=msg)
                msg="%.2f" % (layer[p1]["error"])
                self._printer.update_value(layer_number=i, measurement=dataset["measurement"], column=p2+'_ERR', value=msg)
                
            invalid_values=[]
            invalid_errors=[]
            invalid_ranges=[]
            
            for param in translation_table.keys():
                if layer[param]["nan_flag"] == True:
                    if param != "dep":
                        invalid_values.append(translation_table[param])
                
                if layer[param]["error_flag"] == True:
                    invalid_errors.append(translation_table[param])
                    
                if layer[param]["range_flag"] == True:
                    invalid_ranges.append(translation_table[param])
                
            if len(invalid_values) > 0:
                self._printer.update_value(layer_number=i, measurement=dataset["measurement"], value="Typing not possible: intensive parameter(s) %s cannot be calculated. " % str(invalid_values).replace(',', ';'))
                warningCount += 1
                warningMsg += "Typing not possible in layer %d: intensive parameter(s) %s cannot be calculated. " % (i+1, str(invalid_values).replace(',', ';'))
                continue
                # return RES_ERROR, "Typing not possible: intensive parameter(s) %s cannot be calculated. " % str(invalid_values).replace(',', ';')
                
            if len(invalid_ranges) > 0:
                self._printer.update_value(layer_number=i, measurement=dataset["measurement"], value="Typing not possible: values of the intensive parameter(s) %s out of acceptable range. " % str(invalid_ranges).replace(',', ';'))
                warningCount += 1
                warningMsg += "Typing not possible in layer %d: values of the intensive parameter(s) %s out of acceptable range. " % (i+1, str(invalid_ranges).replace(',', ';'))
                continue
                # return RES_ERROR, "Typing not possible: values of the intensive parameter(s) %s out of acceptable range. " % str(invalid_ranges).replace(',', ';')
            
            if len(invalid_errors) > 0:
                self._printer.update_value(layer_number=i, measurement=dataset["measurement"], value="Typing uncertain: relative error of intensive parameter(s) %s higher than 20%%. " % str(invalid_errors).replace(',', ';'))
                warningCount += 1
                warningMsg += "Typing uncertain: relative error of intensive parameter(s) %s higher than %d%%. " % ( str(invalid_errors).replace(',', ';'), NtGlobals.MAX_ACCEPTED_ERROR_INT * 100 )
                
                wx.PostEvent(
                    self._parent,
                    DatasetWarning(
                        data={
                            'dataset': dataset["measurement"],
                            'warning': "Typing uncertain: relative error of intensive parameter(s) %s higher than %d%%. " % ( str(invalid_errors).replace(',', ';'), NtGlobals.MAX_ACCEPTED_ERROR_INT * 100 )},
                        thread=threading.current_thread()
                    )
                )
                
        if self.Stopped():
            return RES_ABORTED, 'Processing aborted!'
                
        # Use neural networks to provide aerosol typing.
        TP.PrintLine("Interogating neural networks...")
        nn=NeuralProcessing(folder=self._ann_folder, thread=self)
        try:
            votes, stdout, stderr=nn.process_measurement(measurement, self._min_accepted_confidence, self._finesse)
        except WindowsError:
            TP.Log('Typing error. Possibly invalid ANN folder.')
            TP.PrintLine('Typing error. Possibly invalid ANN folder.')
            return RES_ERROR, 'Typing error. Possibly invalid ANN folder.'
        except RuntimeError:
            TP.Log('Processing aborted!')
            TP.PrintLine('Processing aborted!')
            return RES_ABORTED, 'Processing aborted!'
            
        net_output=""
            
        if len(stderr) > 0:
            TP.PrintLine("One or more networks encountered errors! Check report file.")
            TP.Log("One or more networks encountered errors! Check report file.")
            warningCount += 1
            warningMsg += "One or more networks encountered errors! Check log file. "
            
            wx.PostEvent(
                self._parent,
                DatasetWarning(
                    data={
                        'dataset': dataset["measurement"],
                        'warning': "One or more networks encountered errors! Check log file. "
                    },
                    thread=threading.current_thread()
                )
            )
            
            self._printer.measurement_header(measurement=dataset["measurement"], message="Network errors:\n")
            for network, error in stderr.items():
                TP.PrintLine("\t%s: '%s'" % (network, error))
                self._printer.measurement_header(measurement=dataset["measurement"], message="\t%s: %s" % (network, error))
                
        TP.PrintLine("\nDoing election based on network answers...\n")
        
        # Have an election based on the networks' output:
        results = Election(votes=votes, min_ratio=self._min_agreement_percentage).results()
        
        allNA = True
        
        for i in range(0, len(results)):
            layer=results[i]
            
            cr_355_532 = float (self._printer.get_value ( dataset["measurement"], i, "CR355_532" ))
            cr_532_1064 = float (self._printer.get_value ( dataset["measurement"], i, "CR532_1064" ))
            lr_355 = float (self._printer.get_value ( dataset["measurement"], i, "LR355" ))
            lr_532 = float (self._printer.get_value ( dataset["measurement"], i, "LR532" ))
            ae_355_532 = float (self._printer.get_value ( dataset["measurement"], i, "AE355_532" ))
            
            if layer['L_result'] == "Marine/CC":
                if cr_355_532 / cr_532_1064 > 1:
                    layer['L_result'] = "Marine"
                elif lr_355 / lr_532 < 1:
                    layer['L_result'] = "Cloud Corrupted"
                elif ae_355_532 < 1:
                    layer['L_result'] = "Mineral (Coarse)"
                else:
                    layer['L_result'] = "Smog"
                    
            if layer['H_result'] == "Marine/CC":
                if cr_355_532 / cr_532_1064 > 1:
                    layer['H_result'] = "Marine"
                elif lr_355 / lr_532 < 1:
                    layer['H_result'] = "Cloud Corrupted"
                elif ae_355_532 < 1:
                    layer['H_result'] = "Mineral (Coarse)"
                else:
                    layer['H_result'] = "Smog"
            
            self._printer.update_value(layer_number=i, measurement=dataset["measurement"], column="Predominant_Component", value=layer['L_result'][0])
            TP.PrintLine("Predominant aerosol in layer %.2fm -> %.2fm: %s" % (layer['bottom'], layer['top'], layer['L_result'][0]))
            self._printer.update_value(layer_number=i, measurement=dataset["measurement"], column="Aerosol_Type", value=layer['H_result'][0])
            TP.PrintLine("Aerosol type in layer %.2fm -> %.2fm: %s\n" % (layer['bottom'], layer['top'], layer['H_result'][0]))

            self._printer.update_value(layer_number=i, measurement=dataset["measurement"], column="Bottom", value=int(layer['bottom']))
            self._printer.update_value(layer_number=i, measurement=dataset["measurement"], column="Top", value=int(layer['top']))
            for column, output in layer['L_result'][1].items():
                self._printer.update_value(layer_number=i, measurement=dataset["measurement"], column=column, value=output)
            for column, output in layer['H_result'][1].items():
                self._printer.update_value(layer_number=i, measurement=dataset["measurement"], column=column, value=output)
                
            aerosol = layer['L_result'][0]
        
            if layer['H_result'][0] != 'Unknown' and layer['H_result'][0] != 'N/A':
                aerosol = layer['H_result'][0]
                
            print "Aerosol type: %s" % aerosol
            
            if aerosol != 'N/A':
                allNA = False
                
        if allNA == True:
            wx.PostEvent(
                self._parent,
                DatasetError(
                    data={'dataset': dataset["measurement"], 'error': 'Couldn\'t obtain results for any layer!'},
                    thread=threading.current_thread()
                )
            )
                
        wx.PostEvent(
            self._parent,
            AerosolsComputed(
                data={'dataset': dataset["measurement"], 'results': results, 'altitudes': altitudes},
                thread=threading.current_thread()
            )
        )
                
        if warningCount > 0:
            return RES_WARNING, warningMsg
            
        return RES_OK, ""
        
class LidarMeasurement:
    """
    Hold information about a complete LIDAR measurement and
    expose methods to process the data.
    """
            
    _extparams={}
    _intparams={}
        
    def __init__(self, dataset):
        """
        Construct a LidarMeasurement object.
        
        Keyword arguments:
        name   -- the name of the measurement
        folder -- the folder in which the measurement files
                  are located.
        """
        self._name=dataset["measurement"]
        self._listedFiles=dataset["files"]
        self._depolarisation_channel={ "input": "", "output": "" }
        self._format = dataset["format"]
        
        self._data_files = {
            "e355": "",
            "e532": "",
            "b355": "",
            "b532": "",
            "b1064": "",
            "dep": ""
        }
        
        self._extparams = {}
        self._intparams = {}
        
        self._avg_extparams = []
        self._avg_intparams = []
        
    def has_depolarisation(self):
        """
        Check if this measurement contains a depolarisation channel.
        """
        return ("dep" in self._data_files.keys())
     
    def read_data(self):
        """
        Read the actual measurement data.
        """
        # file_list = [ f for f in listdir(self._folder) if re.search(r'(%s).*$'%self._name, f)]
        required_channels = [ 'b355', 'b532', 'b1064', 'e355', 'e532' ]
        
        altitude_channel = 'Altitude'
        channel_names = {}
        
        for c in required_channels:
            channel_names[c] = ""
            
        print ( self._format )
        print ( [ ( channel, os.path.basename(fn) ) for channel, fn in self._listedFiles.items() ] )
        
        for channel, file_name in self._listedFiles.items():
            pretty_fn = os.path.basename ( file_name )
            
            wavelength = channel[1:]
                
            try:
                root = Dataset ( file_name, 'r' )
                vars = root.variables
                
                print "Channels in %s: %s" % ( pretty_fn, vars.keys() )
                
                for ncChannel in vars.keys():
                    # print (ncChannel.lower())
                    
                    if ncChannel.lower() == 'backscatter':
                        # For the backscatter channels, try to read them from the
                        # extiction files, if possible. 
                        channelKey = 'b' + wavelength
                        if channel.startswith ( 'e' ):
                            self._data_files[ channelKey ] = file_name
                            channel_names[ channelKey ] = { "input": ncChannel, "output": "Backscatter_" + wavelength }
                        elif self._data_files[ channelKey ] == '':
                            self._data_files[ channelKey ] = file_name
                            channel_names[ channelKey ] = { "input": ncChannel, "output": "Backscatter_" + wavelength }
                    elif ncChannel.lower() == 'extinction':
                        print ("I am here!")
                        if self._data_files [ 'e' + wavelength ] == "" or channel.startswith ( 'e' ):
                            print ("I am here 2!")
                            self._data_files[ 'e' + wavelength ] = file_name
                            channel_names[ 'e' + wavelength ] = { "input": ncChannel, "output": "Extinction_" + wavelength }
                    elif 'depol' in ncChannel.lower() and ncChannel.lower().startswith( 'error' ) == False and wavelength == '532':
                        # Newer files can contain both particle and volume depolarisation,
                        # but we're only interested in partile depolarisation.
                        # if ncChannel.startswith( 'Vol' ) == False and self._data_files[ 'dep' ] == '':
                        if self._data_files[ 'dep' ] == '':
                            self._data_files [ 'dep' ] = file_name
                            self._depolarisation_channel = { "input": ncChannel, "output": "Depolarisation" }
                    elif ncChannel.lower().startswith('altitude'):
                        altitude_channel = { "input": ncChannel, "output": "Altitude" }
                        
                print ( self._data_files )
                
            except Exception as e:
                print (str(e))
                pass
                
        if self._data_files["dep"] is "":
            del self._data_files["dep"]
                
        missing_files=[]
        for channel in required_channels:
            if self._data_files[ channel ] == '':
                # This file was not found!
                missing_files.append(channel)
        
        if len(missing_files) > 0:
            # One or more required files are missing, oops!
            raise ValueError("Required signals %s are missing!" % missing_files)
        
        unique_files = []
        for file in self._data_files.values():
            if file != '':
                unique_files.append(file)
            
        unique_files=set(unique_files)
        
        parameters={}
        for file in unique_files:
            # Figure out which files we need to read; we don't want to open a file
            # multiple times if it contains more than one channel, so group the
            # channels by file.
            parameters[file] = []
            if file==self._data_files["e355"]: parameters[file].append(channel_names["e355"])
            if file==self._data_files["e532"]: parameters[file].append(channel_names["e532"])
            if file==self._data_files["b355"]: parameters[file].append(channel_names["b355"])
            if file==self._data_files["b532"]: parameters[file].append(channel_names["b532"])
            if file==self._data_files["b1064"]: parameters[file].append(channel_names["b1064"])
            if self.has_depolarisation():
                # Also check for the depolarisation channel.
                if file==self._data_files["dep"]: parameters[file].append(self._depolarisation_channel)
            
        del unique_files
        
        altitudes={}
        data={}
        errors={}
        
        for file_name, parameters in parameters.items():
            # Figure out which channels to read from this file
            to_read = parameters + [altitude_channel]
            # Read those parameters and the altitude series
            # print os.path.join ( self._folder, file_name ), to_read
            data_from_file = read_lidar_file(file_name, to_read)
            
            # Save the altitude series for each file.
            # We will use this later to project the data against
            # a single altitude series.
            altitudes[file_name]=data_from_file[altitude_channel["output"]]["data"]
            
            for parameter, array in data_from_file.items():
                if parameter is not "Altitude":
                    # Skip the altitude series because we
                    # already stored its data
                    data[parameter]=[]
                    errors[parameter]=[]
                    
                    for i in range(0, len(array["data"])):
                        d=array["data"][i]
                        e=array["error"][i]
                        
                        # Check each data point for sanity.
                        # Not all NetCDF files are equal... some were
                        # unfortunate with their maker
                        if parameter == self._depolarisation_channel["output"]:
                            if (array["data"][i] < 0 or array["data"][i] > 0.6):
                                d=np.nan
                                e=np.nan
                        else:
                            if (array["data"][i] < np.power(10, -12) or array["data"][i] > 0.01):
                                d=np.nan
                                e=np.nan
                                
                        if array["data"][i] is ma.masked:
                            # Some NetCDF files have masked data; we should
                            # conver this to NaNs to have some consistency.
                            d=np.nan
                            e=np.nan
                            
                        data[parameter].append(d)
                        errors[parameter].append(e)
        
        # Project every channel against the 1064nm Backscatter file
        # altitude series, because this is what we're going to use
        # to determine the layers.
        #
        # Since we're going to do the processing in-layer, data outside
        # the layers won't matter at all.
        self._altitudes=altitudes[self._data_files['b1064']]
        self._altitudes=[]
        
        for i in range(0, len(altitudes[self._data_files['b1064']])):
            if np.isnan(data['Backscatter_1064'][i]) != True:
                # Some NetCDF files have NaNs in the altitude series (why?!).
                # We're going to simply ignore these data points.
                self._altitudes.append(altitudes[self._data_files['b1064']][i])
        
        td={
            "e355": "Extinction_355",
            "e532": "Extinction_532",
            "b355": "Backscatter_355",
            "b532": "Backscatter_532",
            "b1064": "Backscatter_1064",
            "dep": self._depolarisation_channel["output"]
        }

        for parameter in self._data_files.keys():
            # For each parameter initialize the value and error
            # arrays with NaN values.
            self._extparams[parameter] = {
                "data": [np.nan] * len(self._altitudes),
                "error": [np.nan] * len(self._altitudes)
            }
            
            for i in range(0, len(altitudes[self._data_files[parameter]])):
                # Determine the corresponding array index based on the altitude
                alt=altitudes[self._data_files[parameter]][i]
                ialt=alt2index(self._altitudes, [(alt, alt)])
                
                if len(ialt) > 0:
                    # And copy the value and the error from the original array
                    # to the new one, which is plotted against the "global"
                    # altitude series.
                    altindex=ialt[0][0]
                    if np.isnan(data[td[parameter]][i]) != True:
                        if parameter == 'dep':
                            self._extparams[parameter]["data"][altindex]=data[td[parameter]][i]*100
                            self._extparams[parameter]["error"][altindex]=errors[td[parameter]][i]*100
                        else:
                            self._extparams[parameter]["data"][altindex]=data[td[parameter]][i]
                            self._extparams[parameter]["error"][altindex]=errors[td[parameter]][i]
        
        del altitudes, data, errors
        
    def get_layers(self, min_height=300, window_size=211, order=3):
        """
        Find particle layers based on the B1064 channel.
        
        Keywork arguments:
        min_height  -- the minimum height a layer should have in order
                       to be considered
        window_size -- the size of the window used to compute the second
                       derivative for the B1064 channel
        order       -- polynomial order with which to approximate
                       the B1064 signal during computations.
        """
        if len(self._extparams) is 0:
            raise ValueError('Cannot compute layers before reading data.')
            
        # Determine the spatial resolution of this measurement
        # and compute the window size in number of bins.
        alt_resolution=self._altitudes[1]-self._altitudes[0]
        bin_window_size=int(window_size/alt_resolution)
        if bin_window_size % 2 == 0:
            bin_window_size += 1
            
        # Use Savitzky-Golay two times to smooth out the signal
        # and compute the first and second derivatives.
        d_b1064 = savitzky_golay(y=self._extparams['b1064']["data"], window_size=bin_window_size, order=order, deriv=1)
        d2_b1064 = savitzky_golay(y=d_b1064, window_size=bin_window_size, order=order, deriv=1)
        
        del d_b1064
        
        x=self._altitudes
        y=d2_b1064
        
        layers = []
        layer_start = x[0]
        last_sign = 0
        layer_edges = collections.OrderedDict() # [altitude, 0/1] - 0=bottom, 1=top
    
        # We'll look for sign changes in the second derivative
        # data series, so determine the sign of the first value.
        if y[0] < 0:
            last_sign = -1
        else:
            last_sign = 1
        
        # Find all the layer edges and store if they're
        # bottoms or tops.
        for alt, val in zip(x[1:], y[1:]):
            if last_sign is -1:
                if val>=0:
                    # Positive first derivative -> top
                    layer_edges[alt] = 1
                    last_sign = 1
                elif val<0:
                    last_sign = -1
            if last_sign is 1:
                if val<=0:
                    # Negative first derivative -> bottom
                    layer_edges[alt] = 0
                    last_sign = -1
           
        if len(layer_edges) < 1:
            # We couldn't find any layer edges so there
            # are no layers.
            return []
            
        alt, top = layer_edges.popitem(last=False)
        
        layer_start=0

        if top is 1:
            # We've found the top of the PBL.
            # print "Found PBL layer topping out at %fm" % alt
            layers.append((x[0], alt))
        else:
            layer_start=alt
            
        # Look for pairs of bottoms and tops in the
        # layer edges list. Ignore multiple edges of the
        # same type.
        for alt, top in layer_edges.items():
            if top == 0 and layer_start == 0:
                layer_start = alt
            elif top == 1 and layer_start != 0:
                if alt - layer_start >= min_height:
                    layers.append((layer_start, alt))
                    layer_start = 0
        
        return layers
        
    def get_snr_layers(self, layers=None, min_layer_height=200, min_snr=5):
        '''
        Select middle portions of the layers that fullfill
        a minimum signal-to-noise ratio criteria.
        
        Keyword arguments:
        layers           -- list of layers on which to apply
                            the algorithm
        min_layer_height -- the minimum height of the sublayer
                            to be returned. If the criteria isn't
                            fullfilled, this is the height of the
                            resulting sublayer.
        min_snr          -- the minimum signal-to-noise ratio
                            that the sublayer should have.
        '''
        if layers == None:
            return []
            
        if len(self._altitudes) < 2:
            return []
            
        if "ae_355_532" not in self._intparams.keys():
            return []
            
        # Get the spatial resolution of this specific measurement
        k_resolution=(self._altitudes[1]-self._altitudes[0])
        # And compute the minimum array length for that resolution:
        k_min_length=int(np.divide(min_layer_height, k_resolution)) + 1
        
        ilayers=alt2index(self._altitudes, layers)
        
        snr_ilayers=[]
        
        for layer in ilayers:
            bottom=layer[0]
            top=layer[1]
            
            snr_info=(bottom, top, .0)

            while True:
                # The virtual layer is getting too small,
                # stop looking for better SNR in sublayers.
                if top - bottom <= k_min_length:
                    break
                    
                # Compute the mean value and the error
                # for this virtual layer:
                mean, error = compute_layer_average(
                    bottom, top,
                    self._intparams['ae_355_532']['data'],
                    self._intparams['ae_355_532']['error']
                )
                
                # Compute the SNR ratio:
                snr=np.divide(mean, error)
                
                if snr > snr_info[2]:
                    # This is the highest SNR yet - Store it.
                    snr_info=(bottom, top, snr)
                
                if snr > min_snr:
                    # SNR above accepted level, no need to keep looking
                    # for higher SNRs.
                    break
                
                # Keep looking in further sublayers.
                bottom += 1
                top -= 1
            
            snr_ilayers.append((bottom, top))

        # Return the layers as altitudes:
        snr_layers=index2alt(self._altitudes, snr_ilayers)
        
        # print "For layers %s\nfound SNR layers %s" % (layers, snr_layers)
        
        return snr_layers
            
        
    def compute_intensive_parameters(self, layers):
        """
        Compute the intensive parameters from the extensive ones.
        
        Keyword arguments:
        layers -- layers in which to compute the intensive parameters.
        """
        
        # Initialize intensive parameters arrays with NaNs.
        self._intparams['ae_355_532'] = {'data': [np.nan] * len(self._altitudes), 'error': [np.nan] * len(self._altitudes)}
        self._intparams['ci_355_532'] = {'data': [np.nan] * len(self._altitudes), 'error': [np.nan] * len(self._altitudes)}
        self._intparams['ci_532_1064'] = {'data': [np.nan] * len(self._altitudes), 'error': [np.nan] * len(self._altitudes)}
        self._intparams['cr_355_532'] = {'data': [np.nan] * len(self._altitudes), 'error': [np.nan] * len(self._altitudes)}
        self._intparams['cr_532_1064'] = {'data': [np.nan] * len(self._altitudes), 'error': [np.nan] * len(self._altitudes)}
        self._intparams['lr_355'] = {'data': [np.nan] * len(self._altitudes), 'error': [np.nan] * len(self._altitudes)}
        self._intparams['lr_532'] = {'data': [np.nan] * len(self._altitudes), 'error': [np.nan] * len(self._altitudes)}
    
        np.seterr(divide='ignore', invalid='ignore')
    
        for layer in layers:
            # Compute intensive parameters in-layer
            linfo = alt2index(self._altitudes, [layer])[0]
            # Get start- and end-indexes for the layer
            # boundary altitudes.
            startindex = linfo[0]
            endindex = linfo[1]
            r = self.compute_ln(p1="e355", p2="e532", l=layer, lambda1=355.0, lambda2=532.0)
            if len(r) == 2:
                for i in range(startindex, startindex+len(r[0])):
                    self._intparams['ae_355_532']['error'][i] = r[1][i-startindex]
                    if np.isfinite(r[0][i-startindex]):
                        self._intparams['ae_355_532']['data'][i] = r[0][i-startindex]
        
            r = self.compute_ln(p1="b355", p2="b532", l=layer, lambda1=355.0, lambda2=532.0)
            if len(r) == 2:
                for i in range(startindex, startindex+len(r[0])):
                    self._intparams['ci_355_532']['error'][i] = r[1][i-startindex]
                    if np.isfinite(r[0][i-startindex]):
                        self._intparams['ci_355_532']['data'][i] = r[0][i-startindex]
        
            r = self.compute_ln(p1="b532", p2="b1064", l=layer, lambda1=532.0, lambda2=1064.0)
            if len(r) == 2:
                for i in range(startindex, startindex+len(r[0])):
                    self._intparams['ci_532_1064']['error'][i] = r[1][i-startindex]
                    if np.isfinite(r[0][i-startindex]):
                        self._intparams['ci_532_1064']['data'][i] = r[0][i-startindex]
        
            r = self.compute_divide(p1="b355", p2="b532", l=layer)
            if len(r) == 2:
                for i in range(startindex, startindex+len(r[0])):
                    self._intparams['cr_355_532']['error'][i] = r[1][i-startindex]
                    if np.isfinite(r[0][i-startindex]):
                        self._intparams['cr_355_532']['data'][i] = r[0][i-startindex]
        
            r = self.compute_divide(p1="b532", p2="b1064", l=layer)
            if len(r) == 2:
                for i in range(startindex, startindex+len(r[0])):
                    self._intparams['cr_532_1064']['error'][i] = r[1][i-startindex]
                    if np.isfinite(r[0][i-startindex]):
                        self._intparams['cr_532_1064']['data'][i] = r[0][i-startindex]
        
            r = self.compute_divide(p1="e355", p2="b355", l=layer)
            if len(r) == 2:
                for i in range(startindex, startindex+len(r[0])):
                    self._intparams['lr_355']['error'][i] = r[1][i-startindex]
                    if np.isfinite(r[0][i-startindex]):
                        self._intparams['lr_355']['data'][i] = r[0][i-startindex]
        
            r = self.compute_divide(p1="e532", p2="b532", l=layer)
            if len(r) == 2:
                for i in range(startindex, startindex+len(r[0])):
                    self._intparams['lr_532']['error'][i] = r[1][i-startindex]
                    if np.isfinite(r[0][i-startindex]):
                        self._intparams['lr_532']['data'][i] = r[0][i-startindex]
        
    def average_intparams(self, averaging_layers, layers):
        """
        Average the intensive parameters.
        
        Keyword arguments:
        layers -- layers in which to perform the computations.
        """
        self._avg_intparams=[]
        
        if len(averaging_layers) != len(layers):
            raise ValueError("Number of averaging layers and layer labels arrays must be equal.")
        
        for i in range(0, len(layers)):
            label_bottom = layers[i][0]
            label_top = layers[i][1]
            
            layer = averaging_layers[i]
            linfo = alt2index(self._altitudes, [layer])[0]
            
            bottom = layer[0]
            top = layer[1]
            
            temp_avgs={}
            
            for param in self._intparams.keys():
                # Calculate layer average values and average errors
                # for each intensive parameter.
                temp_avgs[param]={}
                temp_avgs[param]['average'], temp_avgs[param]['error'] = compute_layer_average(
                    start=linfo[0],
                    end=linfo[1],
                    errors=self._intparams[param]['error'],
                    data=self._intparams[param]['data']
                )
                
            for param in ['ae_355_532', 'ci_355_532', 'ci_532_1064', 'cr_355_532', 'cr_532_1064']:
                # Check for invalid values and set the appropiate flags.
                avg=temp_avgs[param]["average"]
                err=temp_avgs[param]["error"]
                temp_avgs[param]["range_flag"]=(avg<-2 or avg>6)
                temp_avgs[param]["nan_flag"]=(np.isfinite(avg) != True or np.isfinite(err) != True)
                temp_avgs[param]["error_flag"]=(np.divide(err, avg) > NtGlobals.MAX_ACCEPTED_ERROR_INT)
                
            for param in ['lr_355', 'lr_532']:
                # Check for invalid values and set the appropiate flags.
                avg=temp_avgs[param]["average"]
                err=temp_avgs[param]["error"]
                temp_avgs[param]["range_flag"]=(avg<5 or avg>200)
                temp_avgs[param]["nan_flag"]=(np.isfinite(avg) != True or np.isfinite(err) != True)
                temp_avgs[param]["error_flag"]=(np.divide(err, avg) > NtGlobals.MAX_ACCEPTED_ERROR_INT)
            
            # Follow the same steps for the depolarisation channel
            # if available.
            
            dep_avg, dep_err = np.nan, np.nan
            
            if self.has_depolarisation():
                dep_avg, dep_err = compute_layer_average(
                    start=linfo[0],
                    end=linfo[1],
                    data=self._extparams['dep']['data'],
                    errors=self._extparams['dep']['data']
                )
                
                temp_avgs["dep"]={}
                temp_avgs["dep"]["average"]=dep_avg
                temp_avgs["dep"]["error"]=dep_err
                temp_avgs["dep"]["range_flag"]=(dep_avg<0 or dep_avg>50)
                temp_avgs["dep"]["nan_flag"]=(np.isfinite(dep_avg) != True or np.isfinite(dep_err) != True)
                temp_avgs["dep"]["error_flag"]=(np.divide(dep_err, dep_avg) > NtGlobals.MAX_ACCEPTED_ERROR_EXT)
                
            temp_avgs["bottom"]=label_bottom
            temp_avgs["top"]=label_top
                
            self._avg_intparams.append(temp_avgs)
                
        return self._avg_intparams
            
    def compute_ln(self, p1, p2, l, lambda1=0.0, lambda2=0.0):
        """
        Compute ln(x/y)/ln(lambda2/lambda1).
        
        Keyword arguments:
        p1      -- first operand
        p2      -- second operand
        l       -- layer in which to perform computation
        lambda1 -- wavelength of the first channel
        lambda2 -- wavelength of the second channel
        """
        result=([], [])
        
        e1=self._extparams[p1]
        e2=self._extparams[p2]
        
        layer=alt2index(self._altitudes, [l])
        start=layer[0][0]
        end=layer[0][1]

        if lambda1 != 0.0 and lambda2 != 0.0:
            c = float(log(lambda2/lambda1))
        else:
            c = 1.0

        for i in range (start, end):                
            x = np.float64(e1["data"][i])
            y = np.float64(e2["data"][i])
            ex = np.float64(e1["error"][i])
            ey = np.float64(e2["error"][i])
            
            if isnan(x) or isnan(y) or isnan(ex) or isnan(ey):
                result[0].append(np.nan)
                result[1].append(np.nan)
                continue
    
            val = np.log(np.divide(x, y))
            val = np.divide(val, c)
            err = np.power(ey, 2)
            err *= np.power(np.divide(1, y), 2)
            err += np.power(ex, 2) * np.power(np.divide(1, x), 2)
            err = np.sqrt(err)
            err *= np.divide(1, c)
            
            result[0].append(val)
            result[1].append(err)
        
        return result

        
    def compute_divide(self, p1, p2, l):
        """
        Compute x/y.
        
        Keyword arguments:
        p1      -- first operand
        p2      -- second operand
        l       -- layer in which to perform computation
        """
        result=([], [])
        
        e1=self._extparams[p1]
        e2=self._extparams[p2]
        
        layer=alt2index(self._altitudes, [l])
        start=layer[0][0]
        end=layer[0][1]

        for i in range (start, end):
            x = np.float64(e1["data"][i])
            y = np.float64(e2["data"][i])
            ex = np.float64(e1["error"][i])
            ey = np.float64(e2["error"][i])
            
            if isnan(x) or isnan(y) or isnan(ex) or isnan(ey):
                result[0].append(np.nan)
                result[1].append(np.nan)
                continue
    
            val = np.divide(x, y)
            err = np.power(ey, 2)
            err *= np.power(np.divide(x, y), 2)
            err += np.power(ex, 2)
            err = np.sqrt(err)
            err = np.divide(err, y)

            result[0].append(val)
            result[1].append(err)
        
        return result
            
    def get_name(self):
        return self._name
        
    def get_folder(self):
        return self._folder
        
    def get_altitudes(self):
        return self._altitudes
        
    def get_extparams(self):
        return self._extparams
        
    def get_intparams(self):
        return self._intparams
        
    def get_avg_intparams(self):
        return self._avg_intparams
        
    def populate_from_array(self, dataset, array):
        """
        Populate the LidarMeasurement object's intensive parameters array
        from an array as returned by the read_intparams_file() method.
        
        Keyword arguments:
        dataset -- name of the measurement.
        array            -- the array holding information about the intensive parameters data.
        """
        self._name = dataset
        
        for layer in array:
            temp_avgs={
                'bottom': layer['bottom'],
                'top': layer['top']
            }
            for param in ['ae_355_532', 'ci_355_532', 'ci_532_1064', 'cr_355_532', 'cr_532_1064']:
                # Check for invalid values and set the appropiate flags.
                avg=layer[param]
                err=layer[param+'_err']
                
                temp_avgs[param]={}
                temp_avgs[param]['average']=avg
                temp_avgs[param]['error']=err
                temp_avgs[param]["range_flag"]=(avg<-2 or avg>6)
                temp_avgs[param]["nan_flag"]=(np.isfinite(avg) != True or np.isfinite(err) != True)
                temp_avgs[param]["error_flag"]=(np.divide(err, avg) > NtGlobals.MAX_ACCEPTED_ERROR_INT)
                
            for param in ['lr_355', 'lr_532']:
                # Check for invalid values and set the appropiate flags.
                avg=layer[param]
                err=layer[param+'_err']
                
                temp_avgs[param]={}
                temp_avgs[param]['average']=avg
                temp_avgs[param]['error']=err
                temp_avgs[param]["range_flag"]=(avg<5 or avg>200)
                temp_avgs[param]["nan_flag"]=(np.isfinite(avg) != True or np.isfinite(err) != True)
                temp_avgs[param]["error_flag"]=(np.divide(err, avg) > NtGlobals.MAX_ACCEPTED_ERROR_INT)
            if 'dep' in layer.keys():
                dep_avg=layer['dep']
                dep_err=layer['dep_err']
                
                temp_avgs["dep"]={}
                temp_avgs["dep"]={}
                temp_avgs["dep"]["average"]=dep_avg
                temp_avgs["dep"]["error"]=dep_err
                temp_avgs["dep"]["range_flag"]=(dep_avg<0 or dep_avg>50)
                temp_avgs["dep"]["nan_flag"]=(np.isfinite(dep_avg) != True or np.isfinite(dep_err) != True)
                temp_avgs["dep"]["error_flag"]=(np.divide(dep_err, dep_avg) > NtGlobals.MAX_ACCEPTED_ERROR_EXT)
                
                # Add a fake depolarisation file to make has_depolarisation() work.
                self._data_files["dep"]="something we don't care about"
            
            self._avg_intparams.append(temp_avgs)
        
'''
End of LidarMeasurement class.
-------------------------------------
Start of various auxiliary functions.
'''

def read_lidar_file (file, measurements):
    '''
    Read LiDAR data from a NetCDF file.
    
    Keyword arguments:
    file         -- path of the NetCDF file.
    measurements -- list of variables to read
                    from the NetCDF files.
    '''
    results = {}
    
    try:
        root = Dataset(file, "r")
        
        for measurement in measurements:
            try:
                varname = measurement["input"][:measurement["input"].index("_")]
            except ValueError:
                # String does not contain "_"!
                varname = measurement["input"]
                
            # print "Trying to read \"%s\" from file \"%s\"" % ( varname, file )
            # print "Available: %s" % root.variables.keys()
            
            data = root.variables[varname][:]
            
            if root.variables[varname].ndim > 1:
                data = np.reshape ( data, root.variables[varname].size )
            
            error = [NtGlobals.MAX_ACCEPTED_ERROR_EXT * d for d in data]
            
            if varname.lower().startswith('alti') != False:
                # Read the error series for everything except Altitude series.
                try:
                    # Of course, the error series might not exist in older files.
                    error = root.variables["Error" + varname][:]
                except KeyError:
                    # If that's the case, presume a constant 20% error.
                    pass
                    
            results[ measurement["output"] ] = {
                "data": data,
                "error": error
            }
            
        root.close()
    except Exception as e:
        raise Exception
    
    return results
    
def read_intparams_file (filename):
    '''
    Read intensive parameters data from a tab-separated text file.
    
    Keyword arguments:
    filename         -- path of the text file.
    '''
    measurements={}
    with open(filename, 'r') as file:
        columns=[]
        line_count = -1 
        
        for line in file:
            line_count += 1
            if line_count == 0:   
                for header in line.split():
                    columns.append(header)
                    
                continue
                
            measurement_data=line.split()
            
            column_count=0
            ldata={}
            for value in line.split():
                if (columns[column_count]) in ['measurement']:
                    ldata[columns[column_count]]=value
                else:
                    ldata[columns[column_count]]=np.float64(value)
                column_count += 1
                
            dataset = ldata['measurement']
            del ldata['measurement']
            
            if dataset not in measurements.keys():
                measurements[dataset]=[]
            
            measurements[dataset].append(ldata)

    return measurements
    
def alt2index(altitudes, layers):
    '''
    Convert altitudes from metres to indexes.
    
    Keyword arguments:
    altitudes -- list of altitudes in metres.
                 The result of this method will
                 consist of indexes of this list.
    layers    -- list of layers to be converted
                 from tuples of altitudes in metres
                 to tuples of altitudes in indexes.
    '''
    indexes = []
    i = 0
    j = 0
    flag = 0

    for bottom, top in layers: 
        while i < len(altitudes) and altitudes[i] < bottom:
            i=i+1
    
        if i >= len(altitudes):
            continue
        
        j=i
    
        while j < len(altitudes) and altitudes[j] < top:
            j=j+1
        
        if j >= len(altitudes):
            continue
        
        indexes.append((i, j))

    return indexes

def index2alt(altitudes, indexes):
    '''
    Convert altitudes from indexes to metres.
    
    Keyword arguments:
    altitudes -- list of altitudes in metres.
                 The result of this method will
                 consist of altitudes from this list.
    layers    -- list of layers to be converted
                 from tuples of altitudes in indexes
                 to tuples of altitudes in metres.
    '''
    layers = []

    for bottom, top in indexes:
        if top < len(altitudes):
            layers.append((altitudes[bottom], altitudes[top]))

    return layers
    
def savitzky_golay(y, window_size, order, deriv=0, rate=1):
    from math import factorial

    try:
        window_size = np.abs(np.int(window_size))
        order = np.abs(np.int(order))
    except ValueError, msg:
        raise ValueError("window_size and order have to be of type int")
    if window_size % 2 != 1 or window_size < 1:
        raise TypeError("window_size size must be a positive odd number")
    if window_size < order + 2:
        raise TypeError("window_size is too small for the polynomials order")
    
    order_range = range(order+1)
    half_window = (window_size-1)/2

    # Precompute coefficients:
    b = np.mat([[k**i for i in order_range] for k in range(-half_window, half_window+1)])
    m = np.linalg.pinv(b).A[deriv] * rate**deriv * factorial (deriv)

    # Pad the signal at the extremes with
    # values taken from the signal itself
    firstvals = y[0] - np.abs(y[1:half_window+1][::-1] - y[0])
    lastvals = y[-1] + np.abs(y[-half_window-1:-1][::-1] - y[-1])
    y = np.concatenate((firstvals, y, lastvals))
    return np.convolve (m[::-1], y, mode='valid')
    
def compute_layer_average(start, end, data, errors):
    '''
    Calculate average value and average error inside
    a layer.
    
    Keyword arguments:
    start  -- first index of data to include in the
              calculation
    end    -- last index of data to include in the
              calculation
    data   -- array of floats representing data point
              values
    errors -- array of floats representing data point
              errors
    '''
    sd=0.0
    se=0.0
    cd=0
    ce=0
    
    for i in range (start, end):
        if np.isfinite(data[i]):
            sd += data[i]
            cd += 1
        if np.isfinite(errors[i]):
            se += np.power(errors[i], 2)
            ce += 1
            
            
    mean = np.divide(sd, cd)
    
    error = np.sqrt(se)
    error = np.divide(error, ce)

    return mean, error
    
def pick_layers(data, layers, min_average=0.0000000250):
    '''
    Filters out layers with the average value
    below a certain threshold.
    data        -- data series to be filtered against
    layers      -- list of layers as tuples containing
                   indexes at the start and the end of
                   the layer, corresponding to the data
                   series
    min_average -- the average value threshold
    '''
    new_layers = []
    
    for bottom, top in layers:
        total = 0
        for value in data[bottom:top]:
            total = total + value
        if total/(top-bottom) > min_average:
            # this layer fullfills the minimum average criteria
            new_layers.append((bottom, top))
    
    return new_layers
    
        