from os import listdir
from os.path import isfile, join, basename
import re
import os

import glob
from datetime import datetime

from netCDF4 import Dataset

APPLICATION_VERSION="1.4.1"
APPLICATION_TITLE="Natali v%s" % APPLICATION_VERSION

LOG_FILENAME='natali_log.txt'

ANN_FOLDER="ANNs"
FILTER_WINDOW=700
MIN_LAYER_DEPTH=300
FINESSE=20
MIN_ACCEPTED_CONFIDENCE=.7
MIN_AGREEMENT_PERCENTAGE=.25
ALTITUDE_SCALE_LOW=0
ALTITUDE_SCALE_HIGH=5

MAX_ACCEPTED_ERROR_EXT=.2
MAX_ACCEPTED_ERROR_INT=.5

AEROSOL_COLORS={
    'Coastal/CC': '#00ffff',
    'CoastalPolluted/CC': '#ff00ff',
    'Continental': '#835c3b',
    'ContinentalDust': '#ffff00',
    'ContinentalPolluted': '#808080',
    'ContinentalSmoke': '#808000',
    'Dust': '#ffa500',
    'DustPolluted': '#ff0000',
    'Marine/CC': '#0000ff',
    # 'MarineMineral/CC': '#008080',
    'MixedDust/CC': '#ffc0cb',
    'MixedSmoke/CC': '#005a00',
    'Smoke': '#00ff00',
    'MineralMixtures/Volcanic': '#9b0022',
    'Marine': '#000000',
    'Cloud Corrupted': '#000000',
    'Mineral (Coarse)': '#000000',
    'Smog': '#000000',
    'N/A': '#eaeaea',
    'Unknown': "#cccccc"
}

FORMAT_UNKNOWN_NETCDF = -1
FORMAT_SCC = 0
FORMAT_OLD_EARLINET_NETCDF = 1
FORMAT_NEW_EARLINET_NETCDF = 2
FORMAT_NEW_EARLINET_NETCDF_3D = 3

NAMING_UNKNOWN = -1
NAMING_SCC = 0
NAMING_OLD_EARLINET = 1
NAMING_NEW_EARLINET = 2

def FindMeasurements (folder):
    '''
    Find measurements inside a given folder.
    
    Keyword arguments:
    folder -- Folder to search for measurements.
    '''
    file_list = [f for f in listdir(folder) if re.search(r'(\.e|\.b)*$', f)]
    
    # file_list = glob.glob(join(folder, '*.nc'))
    # file_list = glob.glob(join(folder, '*'))
    
    groups = []
    
    for file_name in file_list:
        if file_name.endswith('.nc'):
            # Discard extension
            measurement_name = file_name[:file_name[:-3].rfind(".")]
        else:
            measurement_name = file_name[:file_name.rfind(".")]
            
        if not measurement_name in groups:
            groups.append(measurement_name)
    
    groups.sort()
    return groups
    
def DetectFormat ( file ):
    '''
    Tries to determine the NetCDF format used in a specific file.
    
    Keyword arguments:
    file -- Folder to search for measurements.
    '''
    
    dataset = Dataset ( file, "r" )
    
    # Look for older EARLINET format
    if 'Backscatter' in root.variables.keys() or 'Extinction' in root.variables.keys():
        return FORMAT_OLD_EARLINET_NETCDF
    
    # Look for newer EARLINET format
    if 'backscatter' in root.variables.keys():
        if root.variables['backscatter'].ndim > 1:
            return FORMAT_NEW_EARLINET_NETCDF_3D
        
        return FORMAT_NEW_EARLINET_NETCDF
        
    if 'extinction' in root.variables.keys():
        if root.variables['backscatter'].ndim > 1:
            return FORMAT_NEW_EARLINET_NETCDF_3D
        
        return FORMAT_NEW_EARLINET_NETCDF
        
    return FORMAT_UNKNOWN_NETCDF
    
    
def ComposeSCCMeasurementList ( scc_files ):
    results = {}
    
    for file in scc_files:
        measurement = file["station"] + file["props"]["start"].strftime ( "%Y%m%d%H%M%S" )
        
        product_type_id = {
            "b355": -1,
            "b532": -1,
            "b1064": -1,
            "e355": -1,
            "e532": -1
        }
    
        if measurement not in results.keys():
            results[ measurement ] = {
                "measurement": measurement,
                "files": {},
                "station_code": file["station"],
                "start_date": file["props"]["start"].strftime ( "%Y%m%d%H%M%S" ),
                "end_date": file["props"]["stop"].strftime ( "%Y%m%d%H%M%S" ),
                "format_display": "SCC",
                "format": FORMAT_SCC
            }
            
        if file["props"]["product_type_id"] in [0, 3, 7, 8]:
            if product_type_id["b" + str(file["wavelength"])] < file["props"]["product_type_id"]:
                results[ measurement ][ "files" ][ "b" + str(file["wavelength"]) ] = file["path"]
        
        if file["wavelength"] != 1064 and file["props"]["product_type_id"] in [1, 2]:
            if product_type_id["e" + str(file["wavelength"])] < file["props"]["product_type_id"]:
                results[ measurement ][ "files" ][ "e" + str(file["wavelength"]) ] = file["path"]
            
    return list(results.values())
    
def ComposeOldEARLINETList ( earlinet_files ):
    results = {}
    
    for file in earlinet_files:
        measurement = file["station"] + file["props"]["start"].strftime ( "%Y%m%d%H%M%S" )
    
        if measurement not in results.keys():
            results[ measurement ] = {
                "measurement": measurement,
                "files": {},
                "station_code": file["station"],
                "start_date": file["props"]["start"].strftime ( "%Y%m%d%H%M%S" ),
                "format_display": "EARLINET",
                "format": FORMAT_OLD_EARLINET_NETCDF
            }
        
        key = file["props"]["type"] + str(file["wavelength"])
        results[ measurement ][ "files" ][ key ] = file["path"]
        
        replaced_key = 'e' + str(file["wavelength"])
        
        if file["props"]["type"] == 'b':
            if replaced_key not in results[measurement][ "files" ].keys():
                results[ measurement ][ "files" ][ replaced_key ] = file["path"]
            
    return list(results.values())
    
def ComposeNewEARLINETList ( earlinet_files, minlevel = 2 ):
    results = {}
    
    for file in earlinet_files:
        measurement = file["station"] + file["props"]["start"].strftime ( "%Y%m%d%H%M%S" )
    
        if measurement not in results.keys():
            results[ measurement ] = {
                "measurement": measurement,
                "files": {},
                "station_code": file["station"],
                "start_date": file["props"]["start"].strftime ( "%Y%m%d%H%M%S" ),
                "end_date": file["props"]["stop"].strftime ( "%Y%m%d%H%M%S" ),
                "format_display": "EARLINET",
                "format": FORMAT_NEW_EARLINET_NETCDF
            }
        
        key = file["props"]["type"] + str(file["wavelength"])
        results[ measurement ][ "files" ][ key ] = file["path"]

        replaced_key = 'e' + str(file["wavelength"])
        
        if file["props"]["type"] == 'b':
            if replaced_key not in results[measurement][ "files" ].keys():
                results[ measurement ][ "files" ][ replaced_key ] = file["path"]
            
    return list(results.values())
        
def ReadFolder ( folder ):
    '''
    Find measurements inside a given folder.
    
    Keyword arguments:
    folder -- Folder to search for measurements.
    quick_scan -- If True, will only look at the first NetCDF file and assume homogenity.
        Otherwise, will read all files to determine NetCDF format individually.
    '''
    
    # Filter files by naming format
    scc_pattern = re.compile ("^(.+)_([0-9]+)_([0-9]+)_[0-9]+_([0-9]+)_([0-9]+)_([a-z0-9]+)_[a-z0-9]+_v([0-9\.]+)\.nc$")
    old_earlinet_pattern = re.compile ( "^([a-z]+)([0-9]+)\.([be])([0-9]+)\.nc$" )
    kelly_pattern = re.compile ("^([0-9]+)_([0-9]+)_([0-9]+)_(.+)_(.+)_(.+)_([be])([0-9]+)\.nc$")
    new_earlinet_pattern = re.compile ( "^EARLINET_AerRemSen_([a-z]+)_Lev([0-9]+)_([be])([0-9]+)_([0-9]+)_([0-9]+)_v([0-9]+)_qc([0-9]+)\.nc$" )
    
    files = {
        NAMING_UNKNOWN: [],
        NAMING_SCC: [],
        NAMING_OLD_EARLINET: [],
        NAMING_NEW_EARLINET: []
    }
    
    # Identify all NetCDF files in folder
    # netcdf_files = glob.glob ( os.path.join ( folder, '*.nc' ) )
    
    for dirpath, dirnames, fs in os.walk (folder):
        for file_name in fs:
            if not file_name.endswith ('.nc'):
                continue
                
            netcdf_file = os.path.join (dirpath, file_name)
    
    # for netcdf_file in netcdf_files:
        # Extract file name from path
        # file_name = os.path.basename ( netcdf_file )
        
            # Check if file matches SCC naming format
            match = scc_pattern.search ( file_name )
            
            if match is not None:
                files[ NAMING_SCC ].append ({
                    "path": netcdf_file,
                    "station": match.group(1),
                    "wavelength": int ( match.group (3) ),
                    "props": {
                        "product_type_id": int ( match.group(2) ),
                        "start": datetime.strptime ( match.group (4), "%Y%m%d%H%M" ),
                        "stop": datetime.strptime ( match.group (5), "%Y%m%d%H%M" ),
                        "measurement_id": match.group (6),
                        "elda_version": match.group(7)
                    }
                })
                
                continue
                
            match = kelly_pattern.search ( file_name )
            
            if match is not None:
                files[ NAMING_OLD_EARLINET ].append ({
                    "path": netcdf_file,
                    "station": match.group(4),
                    "wavelength": int ( match.group(8) ),
                    "props": {
                        "start": datetime.strptime ( match.group(1) + match.group(2), "%Y%m%d%H%M" ),
                        "type": match.group(7)
                    }
                })
                
                continue
                
            # Check if file matches older EARLINET naming scheme
            match = old_earlinet_pattern.search ( file_name )
            if match is not None:
                files[ NAMING_OLD_EARLINET ].append ({
                    "path": netcdf_file,
                    "station": match.group(1),
                    "wavelength": int ( match.group (4) ),
                    "props": {
                        "start": datetime.strptime ( match.group (2), "%y%m%d%H%M" ),
                        "type": match.group(3)
                    }
                })
                
                continue
                
            # Check if file matches newer EARLINET naming scheme
            match = new_earlinet_pattern.search ( file_name )
            if match is not None:
                files[ NAMING_NEW_EARLINET ].append ({
                    "path": netcdf_file,
                    "station": match.group(1),
                    "wavelength": int ( match.group (4) ),
                    "props": {
                        "level": int ( match.group (2) ),
                        "start": datetime.strptime ( match.group (5), "%Y%m%d%H%M" ),
                        "stop": datetime.strptime ( match.group (6), "%Y%m%d%H%M" ),
                        "type": match.group(3),
                        "version": match.group (7),
                        "qc": int ( match.group(8) )
                    }
                })
                
                continue
                
            files[ NAMING_UNKNOWN ].append ({
                "path": netcdf_file,
                "wavelength": 0,
                "props": {}
            })
            
    print ( "%d unknown files." % len(files[ NAMING_UNKNOWN ]) )
    
    scc_list = ComposeSCCMeasurementList ( files[ NAMING_SCC ] )
    old_earlinet_list = ComposeOldEARLINETList ( files[ NAMING_OLD_EARLINET ] )
    new_earlinet_list = ComposeNewEARLINETList ( files[ NAMING_NEW_EARLINET ], minlevel = 2 )
    
    final_list = scc_list + old_earlinet_list + new_earlinet_list
    final_list.sort ( key = lambda x: x.get("start_date") )
    
    return final_list