import re
from collections import OrderedDict
import wx

class ResultPrinter:
    '''
    Collect and print measurement results and comments
    to CSV or TXT files.
    '''
    _measurements={}
    # Store the columns that will be printed to the file (and their measurement units)
    # as an OrderedDict to print them in the desired order.
    _valid_columns=OrderedDict([
        ('Bottom', '[m]'), ('Top', '[m]'),
        ('Retrieval Bottom', '[m]'), ('Retrieval Top', '[m]'),
        ('AE355_532', ''), ('AE355_532_ERR', ''),
        ('CI355_532', ''), ('CI355_532_ERR', ''),
        ('CI532_1064', ''), ('CI532_1064_ERR', ''),
        ('CR355_532', ''), ('CR355_532_ERR', ''),
        ('CR532_1064', ''), ('CR532_1064_ERR', ''),
        ('LR355', '[sr]'), ('LR355_ERR', '[sr]'),
        ('LR532', '[sr]'), ('LR532_ERR', '[sr]'),
        ('DEP532', ''), ('DEP532_ERR', ''),
        ('Predominant_Component', ''), ('Aerosol_Type', ''),
        ('Comments', ''),
        ('A1H_Answer', ''), ('A1H_Confidence', ''), ('A1H_Agreements', ''),
        ('A2H_Answer', ''), ('A2H_Confidence', ''), ('A2H_Agreements', ''),
        ('A3H_Answer', ''), ('A3H_Confidence', ''), ('A3H_Agreements', ''),
        ('A1L_Answer', ''), ('A1L_Confidence', ''), ('A1L_Agreements', ''),
        ('A2L_Answer', ''), ('A2L_Confidence', ''), ('A2L_Agreements', ''),
        ('A3L_Answer', ''), ('A3L_Confidence', ''), ('A3L_Agreements', ''),
        ('B1L_Answer', ''), ('B1L_Confidence', ''), ('B1L_Agreements', ''),
        ('B2L_Answer', ''), ('B2L_Confidence', ''), ('B2L_Agreements', ''),
        ('B3L_Answer', ''), ('B3L_Confidence', ''), ('B3L_Agreements', ''),
    ])
    
    _csv_file=None
    _rep_file=None
    _output_started=False
   
    
    def __init__(self, unbuffered=False, csv_file="", rep_file=""):
        '''
        Create a ResultPrinter object.
        
        Keyword arguments:
        unbuffered -- if True the printer will output messages to files
                      as soon as they come. If False, it will require to
                      flush the output using the print_csv() and
                      print_friendly() methods.
        csv_file   -- name of the CSV file to be used. If the printer is
                      unbuffered this argument is required.
        rep_file   -- name of the TXT file to be used. If the printer is
                      unbuffered this argument is required.
        '''
        self._header=""
        self._footer=""
        self._unbuffered=unbuffered
        
        if self._unbuffered==True:
            # If we're going to print without buffering
            # we need to open the CSV and TXT files right away.
            if csv_file=="":
                raise ValueError("Invalid file %s!" % csv_file)
            if rep_file=="":
                raise ValueError("Invalid file %s!" % rep_file)
                
            self._csv_file = open(csv_file, 'w', 0)
            self._rep_file = open(rep_file, 'w', 0)
            
    
    def header(self, message):
        '''
        Append a string to the TXT file header
        section.
        
        Keyword arguments:
        message -- the string to be appended.
        '''
        self._header += message
        
    def footer(self, message):
        '''
        Append a string to the TXT file footer
        section.
        
        Keyword arguments:
        message -- the string to be appended.
        '''
        self._footer += message
        
    def has_measurement(self):
        '''
        Check if the printer has been already
        used for one or more measurements.
        '''
        if len(self._measurements) > 0:
            return True
            
        return False
            
    def init_measurement(self, measurement=""):
        '''
        Add a new measurement to the printer.
        
        Keyword arguments:
        measurement -- name of the measurement.
        '''
        if self._unbuffered == True:
            # If the printer is unbuffered we should flush
            # previously unflushed measurements.
            #
            # This is what an unbuffered ResultPrinter means,
            # after all.
            if self._output_started == False:
                # If this is the first measurement print the header
                # for the CSV and TXT files.
                self._output_started = True
                self._csv_file.write(self.get_csv_header())
                self._rep_file.write(self.get_rep_header())
                
            for name, info in self._measurements.items():
                # Print all remaining measurements.
                if info['flushed'] == False:
                    self._csv_file.write(self.get_csv_text(name, info))
                    self._rep_file.write(self.get_rep_text(name, info))
                    self._measurements[name]['flushed'] = True
            
        if measurement == "":
            raise ValueError('Invalid measurement name')
        
        # Add the new measurement to the printer.
        self._measurements[measurement]={
                'header': '',
                'footer': '',
                'flushed': False,
                'layers': []
        }
        
    def close(self):
        '''
        Close the printer; flush output and close the files.
        '''
        if self._unbuffered == True:
            # We will only close the printer if it's currently
            # working unbuffered.
            #
            # This will make the printer work as buffered.
            self._unbuffered = False
            
            for name, info in self._measurements.items():
                # Print all remaining measurements.
                if info['flushed'] == False:
                    self._csv_file.write(self.get_csv_text(name, info))
                    self._rep_file.write(self.get_rep_text(name, info))
                    self._measurements[name]['flushed'] = True
                    
            self._rep_file.write(self.get_rep_footer())
            
            # Close CSV and TXT files
            self._csv_file.close()
            self._rep_file.close()
            
    def create_layers(self, measurement, number_of_layers=0):
        '''
        Initialize layers for a given measurement for this printer.

        Keyword arguments:
        measurement      -- name of the measurement.
        number_of_layers -- how many layers to create for this layer.
        '''
        for i in range(0, number_of_layers):
            self._measurements[measurement]['layers'].append(
                OrderedDict(zip(self._valid_columns.keys(), ['']*len(self._valid_columns)))
            ) 
        
    def measurement_header(self, measurement="", message=""):
        '''
        Append a string to the TXT file in the
        header section of a measurement.
        
        Keyword arguments:
        measurement -- name of the measurement.
        message     -- the string to be appended.
        '''
        if measurement == "" or measurement not in self._measurements.keys():
            raise ValueError('Invalid measurement name')
            
        self._measurements[measurement]['header'] += message
        
    def measurement_footer(self, measurement="", message=""):
        '''
        Append a string to the TXT file in the
        footer section of a measurement.
        
        Keyword arguments:
        measurement -- name of the measurement.
        message     -- the string to be appended.
        '''
        if measurement == "" or measurement not in self._measurements.keys():
            raise ValueError('Invalid measurement name')
            
        self._measurements[measurement]['footer'] += message
        
    def update_value(self, measurement="", layer_number=-1, column="", value=""):
        '''
        Update value for a column inside a layer for this printer.
        
        Keyword arguments:
        measurement  -- name of the measurement for which to update the layer
        layer_number -- layer index (for this measurement)
        column       -- which field to update
        value        -- new value to be stored
        '''
        
        if measurement == "" or measurement not in self._measurements.keys():
            raise ValueError('Invalid measurement name')
            
        if layer_number < 0 or layer_number >= len(self._measurements[measurement]['layers']):
            # We can't update a non-existing layer
            return
            
        # By default, any output regarding a layer will be considered
        # an observation
        if column == "":
            column = "Comments"
            
        # We can only update certain columns
        if column not in self._valid_columns.keys():
            raise KeyError('Invalid column "%s"!' % column)
            
        if column == "Comments":
            # For the 'Comments' column, append messages
            self._measurements[measurement]['layers'][layer_number]["Comments"] += value
        else:
            # For all the other columns, replace the entire value
            self._measurements[measurement]['layers'][layer_number][column] = value
            
    def get_value (self, measurement="", layer_number=-1, column=""):
        if measurement == "" or measurement not in self._measurements.keys():
            raise ValueError('Invalid measurement name')
            
        if layer_number < 0 or layer_number >= len(self._measurements[measurement]['layers']):
            # We can't update a non-existing layer
            return

        if column not in self._valid_columns.keys():
            raise KeyError('Invalid column "%s"!' % column)
            
        return self._measurements[measurement]['layers'][layer_number][column]
            
    def print_csv(self, filename="", separator=','):
        '''
        Write the CSV file.
        
        Keyword arguments:
        filename  -- name of the file to print the
                     CSV data to.
        separator -- separator to be used for delimiting
                     values inside the file.
        '''
        
        try:
            with open(filename, "w") as file:
                # Write the header
                file.write(self.get_csv_header(separator=separator))
                
                # Write a line for each layer
                # Copy the list so we don't mess with it:
                for name, info in self._measurements.items():
                    file.write(self.get_csv_text(name, info, separator=separator))

        except IOError as e:
            # Something went wrong!
            TextPrinter.Print ("I/O error(%s): %s" % (e.errno, e.strerror))
            raise e
            
    def print_friendly(self, filename=""):
        '''
        Write the TXT file.
        
        Keyword arguments:
        filename -- name of the file to print
                    the output to.
        '''
        try:
            with open(filename, "w") as file:
                file.write(self.get_rep_header())
                
                for name, info in self._measurements.items():
                    file.write(self.get_rep_text(name, info))
                    
                file.write(self.get_rep_footer())
                    
        except IOError as e:
            # Something went wrong!
            TP.PrintLine("I/O error(%s): %s" % (e.errno, e.strerror))
            raise e
            
    def get_csv_header(self, separator=','):
        '''
        Get the CSV columns header.
        
        Keyword arguments:
        separator -- separator to be used for delimiting
                     values inside the file.
        '''
        result = ""
        line=separator.join([x + self._valid_columns[x] for x in self._valid_columns.keys()])
        line=separator.join(('Measurement', line))
        result += line
        result += "\n"
        
        return result
            
    def get_csv_text(self, measurement_name, measurement_info, separator=','):
        '''
        Get the CSV output for the specified measurement.
        
        Keyword arguments:
        measurement_name -- name of the measurement.
        measurement_info -- measurement data.
        separator        -- separator to be used for delimiting
                            values inside the file.
        '''
        
        result = ""
        for layer in measurement_info['layers']:
            text=OrderedDict()
            for column in self._valid_columns:
                # Escape quotes:
                if layer[column] == '':
                    if "_Answer" in column:
                        layer[column]="N/A"
                    elif "_Confidence" in column:
                        layer[column]="0"
                    elif "_Agreements" in column:
                        layer[column]="0"
                    elif "DEP" in column:
                        layer[column]="N/A"
                        
                text[column]=re.sub('"', '""', str(layer[column]))
                # Wrap the whole string in double quotes if the separator
                # is present inside the column value:
                if separator in text[column]:
                    text[column]='"%s"' % text[column]
                    
            # Finally write the line to the file
            line=separator.join(text.values())
            line=separator.join((measurement_name, line))
            result += line
            result += "\n"
            
        return result
            
    def get_rep_header(self):
        '''
        Get the TXT file header.
        '''
        
        result = ""
        
        if len(self._header) > 0:
            return (self._header+"\n\n")
            
        return ""
        
    def get_rep_footer(self):
        '''
        Get the TXT file footer.
        '''
        
        if len(self._footer) > 0:
            return ("\n--------------------------------------\n"+self._footer)
            
        return ""
            
    def get_rep_text(self, measurement_name, measurement_info):
        '''
        Get the TXT output for the specified measurement.
        
        Keyword arguments:
        measurement_name -- name of the measurement.
        measurement_info -- measurement data.
        '''
        
        result=""
        
        padded_name = measurement_name.center(36)
        padded_name="|"+padded_name+"|"
        
        result += "+------------------------------------+\n"
        result += padded_name+"\n"
        result += "+------------------------------------+\n"
        
        if len(measurement_info['header']) > 0:
            result += measurement_info['header']+"\n"
        
        i=1
        
        for layer in measurement_info['layers']:
            result += "Layer %d:\n" % i
            
            for column in self._valid_columns.keys():
                if layer[column] == '':
                    if "_Answer" in column:
                        layer[column]="N/A"
                    elif "_Confidence" in column:
                        layer[column]="0"
                    elif "_Agreements" in column:
                        layer[column]="0"
                    elif "DEP" in column:
                        layer[column]="N/A"
                        
                value=layer[column]
                result += "\t%s: %s %s\n" % (column, value, self._valid_columns[column])
                
            i += 1
        
        result += measurement_info['footer']+"\n"
        
        return result

PrintText, EVT_PRINT_TEXT = wx.lib.newevent.NewEvent()        
        
class TextPrinter:
    parent = None
    logFile = None
    
    @staticmethod
    def InitLogFile(filename):
        TextPrinter.logFile = open(filename, 'a', 0)
    
    @staticmethod
    def Print(text):
        if TextPrinter.parent == None:
            return
            
        wx.PostEvent(
            TextPrinter.parent,
            PrintText(
                data=text
            )
        )
        
    @staticmethod
    def PrintLine(text):
        if TextPrinter.parent == None:
            return
            
        wx.PostEvent(
            TextPrinter.parent,
            PrintText(
                data=str(text)+"\n"
            )
        )
        
    @staticmethod
    def Log(text):
        TextPrinter.logFile.write(text)