import subprocess
import os
import time
import numpy as np
import sys

from nt_output import TextPrinter as TP

class NNVotes:
    '''
    Store the network answers and groups them by
    the networks' typing resolution (H or L).
    '''
    def __init__(self, answers=[]):
        '''
        Construct the NNVotes object.
        
        Keyword arguments:
        answers -- list of answers from the networks
        '''
        self._layers=[]
        for layer in answers:
            result={
                'L_answers': [],
                'H_answers': [],
                'bottom': layer['bottom'],
                'top': layer['top'],
                'depolarisation': layer['depolarisation']
            }
            
            for answers in layer["networks"]:
                if answers["name"].endswith('L'):
                    result['L_answers'].append((answers["name"], answers["result"]))
                else:
                    result['H_answers'].append((answers["name"], answers["result"]))
                
            self._layers.append(result)
            
    def per_layer(self):
        '''
        Get the network votes grouped by layer.
        '''
        return self._layers
        
class NNResult:
    '''
    Store the results from a neural network.
    '''
    _type_num=0
    _answers={}
    _min_prob=.7
    
    def __init__(self, known_aerosol_types=None, min_prob=.7):
        '''
        Construct the NNResult object.
        
        Keyword arguments:
        known_aerosol_types -- how many aerosol types the network
                               can distinguish (required).
        min_prob            -- the minimum confidence an answer
                               needs to have in order to be
                               taken into account
        '''
        
        if known_aerosol_types is None:
            raise ValueError("Unknown number of aerosol types")
            
        _answer_count=0
        self._type_num=known_aerosol_types
        self._min_prob=min_prob
        
        if self._type_num==14:
            self._types=[
                'Coastal/CC', 'CoastalPolluted/CC', 'Continental', 'ContinentalDust',
                'ContinentalPolluted', 'ContinentalSmoke', 'Dust', 'DustPolluted',
                'Marine/CC', 'MineralMixtures/Volcanic', 'MixedDust/CC', 'MixedSmoke/CC', 'Smoke', 'MineralMixtures/Volcanic'
            ]
        elif self._type_num==6:
            self._types=[
                'Continental', 'ContinentalPolluted',
                'Dust', 'Marine/CC', 'Smoke', 'MineralMixtures/Volcanic'
            ]
        elif self._type_num==5:
            self._types=[
                'Continental', 'ContinentalPolluted', 'Dust', 'Marine/CC', 'Smoke'
            ]
            
        self._aerosols={}
        for i in range(0, self._type_num):
            self._aerosols[self._types[i]]={"confidence": .0, "confident_answers": 0, "trust_level": .0}

        self._total_count=0
        self._total_with_agreement=0
        self._prob_sums=[.0]*self._type_num
        self._trust_prob_sums=[.0]*self._type_num
        
    def add(self, answers):
        '''
        Add a new set of answers, as read from the network.
        
        Keyword arguments:
        answers -- this is a list of confidence values. Its
                   size has to be same as the number of aerosol
                   types the network can distinguish.
        '''
        
        if len(answers) != self._type_num:
            # This may be garbage read from the file!
            return
            
        self._total_count+=1
        
        with_agreement = False
        
        for i in range(0, len(answers)):
            
            self._prob_sums[i]+=answers[i]
            
            if answers[i] > self._min_prob:
                self._aerosols[self._types[i]]["confident_answers"] += 1
                self._trust_prob_sums[i] += answers[i]
                with_agreement = True
                
        if with_agreement:
            self._total_with_agreement += 1
                
    def answers(self, confidence_weight=.5):
        '''
        Get the answers given by the network.
        
        Keyword arguments:
        confidence_weight -- weight of the confidence of each answer
                             when calculating the overall trust level
                             of the network.
        '''
        
        k=confidence_weight
        if (k > 1 or k < 0):
            # Weight needs to be in interval (0, 1)
            # Fallback to default value if false.
            k = .5
            
        for i in range(0, self._type_num):
            self._aerosols[self._types[i]]["confidence"]=np.divide(self._trust_prob_sums[i], self._aerosols[self._types[i]]["confident_answers"])
            self._aerosols[self._types[i]]["trust_level"]=k*self._trust_prob_sums[i]+(1-k)*self._aerosols[self._types[i]]["confident_answers"]
            # self._aerosols[self._types[i]]["trust_level"]=k*self._aerosols[self._types[i]]["confidence"]+(1-k)*np.divide(self._aerosols[self._types[i]]["confident_answers"], self._total_with_agreement)

        return self._aerosols
            
    def count(self):
        '''
        Get the total number of confident answers
        given by the network.
        '''
        return self._total_count
              
    def reset(self):
        '''
        Reset everything; this gives the equivalent of a new NNResult object.
        '''
        
        for i in range(0, self._type_num):
            self._aerosols[self._types[i]]={"confidence": .0, "confident_answers": 0, "trust_level": .0}
            
        self._answer_count=0
        self._prob_sums=[.0]*len(self._type_num)
        self._trust_prob_sums[.0]*self._type_num
    
class NeuralNetwork:
    '''
    Store information about an ANN and expose methods
    to interact with them.
    '''
    def __init__(self, name="", folder=""):
        '''
        Construct a NeuralNetwork object.
        
        Keyword arguments:
        name   -- the name of the ANN.
        folder -- the folder in which the ANN files
                  are located.
        '''
        self._name=name
        self._folder=folder
        self._process=None
        self._output=None
        
        if name.startswith('A'):
            if name.endswith('H'):
                self._aerosol_types=14
            elif name.endswith('L'):
                self._aerosol_types=6
        elif name.startswith('B'):
            self._aerosol_types=5
        
        for file in os.listdir(self._folder):
            if file.endswith(".exe"):
                self._binary_name=file
        
    def input_filename(self):
        '''
        Get the full path of the ANN input file.
        '''
        
        return self._folder+"\\inputFile.asc"
        
    def output_filename(self):
        '''
        Get the full path of the ANN output file.
        '''
        
        return self._folder+"\\activeOutputProbe.asc"
        
    def binary_name(self):
        '''
        Get the name of the executable file of the ANN.
        '''
        
        return self._binary_name
        
    def name(self):
        '''
        Get the name of the ANN.
        '''
        
        return self._name
        
    def folder(self):
        '''
        Get the folder in which the ANN's files are located.
        '''
        return self._folder
        
    def running(self):
        '''
        Check if the ANN is running.
        '''
        if self._process is None:
            return False
            
        if self._process.poll() is None:
            return True
        
        return False
        
    def result(self, min_probability):
        '''
        Get the result of the ANN's latest run
        as an NNResult object.
        
        Keyword arguments:
        min_probability -- the minimum confidence an answer needs
                           to have in order for it to be taken
                           into account.
        '''
        parameters=[]
        result=NNResult(known_aerosol_types=self._aerosol_types, min_prob=min_probability)
        
        with open(self._folder+"\\activeOutputProbe.asc", "r") as output_file:
            for line in output_file:
                parameters=[]
                try:
                    parameters=[np.float64(x) for x in line.split()]
                except ValueError:
                    continue
                
                result.add(parameters)
                        
        return result
        
    def stdout(self):
        '''
        Get the standard output of the ANN after its latest run.
        '''
        
        if self._process is None:
            # Check if there's any process to read the output from.
            return ""
            
        if self._process.poll() is None:
            # Check if the process has finished.
            return ""
            
        if self._output is None:
            # Read the output.
            self._output=self._process.communicate()
            
        # Return standard output.
        return self._output[0]
        
    def stderr(self):
        '''
        Get the error output of the ANN after its latest run.
        '''
        if self._process is None:
            # Check if there's any process to read the output from.
            return ""
            
        if self._process.poll() is None:
            # Check if the process has finished.
            return ""
            
        if self._output is None:
            # Read the output.
            self._output=self._process.communicate()

        # Return error output.   
        return self._output[1]
                
    def start(self):
        '''
        Run the ANN executable file.
        '''
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        self._process=subprocess.Popen(self._folder+"\\"+self._binary_name, cwd=self._folder, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE, startupinfo=startupinfo)
        
    def stop(self):
        '''
        Terminate the ANN process.
        '''
        self._process.terminate()
        
        # Also destroy the subprocess
        self._process=None

class NeuralProcessing:
    '''
    Process LiDAR measurements by means of
    neural processing techniques using
    artificial neural networks.
    '''
    _nn_folder=""
    _binaries=[]
    _depolarisation=False
    _networks=[]
    _dep_networks=[]
    _no_dep_networks=[]
    _min_probability=0.7
    
    def __init__(self, thread, folder="F:\\NATALI\\ANNs\\NewANNs"):
        '''
        Construct the NeuralProcessing object.
        
        Keyword arguments:
        folder -- folder containing the ANNs.
        '''
        self._nn_folder=folder
        self._thread = thread
        
    def process_measurement(self, measurement, min_probability=.7, samples_around_mean=50):
        '''
        Process the specified measurement.
        
        Keyword arguments:
        measurement         -- LidarMeasurement object
        min_probability     -- the minimum confidence an ANN answer needs
                               to have in order for it to be taken
                               into account.
        samples_around_mean -- how many values to generate in the
                               [value-uncertainty; value+uncertainty]
                               interval.
        '''
        self._with_depolarisation=measurement.has_depolarisation()
        self._min_probability=min_probability
        
        self._dep_networks=[]
        self._no_dep_networks=[]
        
        self._dep_networks=[NeuralNetwork(name=name, folder=self._nn_folder+"\\"+name) for name in ['A1H', 'A2H', 'A3H', 'A1L', 'A2L', 'A3L']]
        self._no_dep_networks=[NeuralNetwork(name=name, folder=self._nn_folder+"\\"+name) for name in ['B1L', 'B2L', 'B3L']]
        
        avg_intparams=measurement.get_avg_intparams()
        scrambled_intparams = scramble(avg_intparams, samples=samples_around_mean)
        
        answers=[]
        network_stdout={}
        network_stderr={}
    
        for layer in scrambled_intparams:
            # Process layer by layer
            TP.PrintLine("Processing layer %.0fm -> %.0fm" % (layer['bottom'], layer['top']))
            
            network_answers=[]
            input_data=""
            has_dep=True
            
            
            for i in range (len(layer["parameters"])):
                if ( 'dep' in layer["parameters"][i].keys() ):
                    del layer["parameters"][i]['dep']
            
            if ('dep' in layer["parameters"][0].keys() and
                 np.isfinite(layer["parameters"][0]['dep'])):
                # Depolarisation channel is available in this layer
                for ps in layer["parameters"]:
                    everything_ok=True
                    for param, value in ps.items():
                        if np.isfinite(value) != True:
                            everything_ok = False
                        
                    if everything_ok != True:
                        # Add parameter set to the input data
                        # only if all parameter values are valid.
                        continue
                        
                    # Generate input.asc
                    # AE350_550, CI350_550, CI550_1000, CR350_550, CR350_1000, LR350, LR550, DEP550
                    input_data+="%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\n" % (
                        ps["ae_355_532"],
                        ps["ci_355_532"], ps["ci_532_1064"],
                        ps["cr_355_532"], ps["cr_532_1064"],
                        ps["lr_355"], ps["lr_532"],
                        ps["dep"]
                    )
                    
                if input_data != "":
                    # Run the networks only if we could generate the input data.
                    network_answers=self.run_networks(self._dep_networks, input_data)
                    
                    for network in self._dep_networks:
                        # Read each network's output, but only store
                        # it if there's anything relevant.
                        stdout = network.stdout()
                        stderr = network.stderr()
                        
                        if len(stdout) > 0:
                            network_stdout[network.name()] = stdout
                            
                        if len(stderr) > 0:
                            network_stderr[network.name()] = stderr
                else:
                    # We couldn't generate the input data for the ANN
                    # so there's no point in further processing.
                    TP.PrintLine("\tSkipping!")
            else:
                # Depolarisation channel is not available in this layer
                has_dep=False
                for ps in layer["parameters"]:
                    everything_ok=True
                    for key, value in ps.items():
                        if key == 'dep':
                            # Don't check the depolarization channel
                            # since we're going to use the B class
                            # ANNs anyway
                            continue
                            
                        if np.isfinite(value) != True:
                            everything_ok = False
                        
                    if everything_ok != True:
                        # Add parameter set to the input data
                        # only if all parameter values are valid.
                        continue
                        
                    # Generate input.asc
                    # AE350_550, CI350_550, CI550_1000, CR350_550, CR350_1000, LR350, LR550
                    input_data+="%f\t%f\t%f\t%f\t%f\t%f\t%f\n" % (
                        ps["ae_355_532"],
                        ps["ci_355_532"], ps["ci_532_1064"],
                        ps["cr_355_532"], ps["cr_532_1064"],
                        ps["lr_355"], ps["lr_532"]
                    )
                    
                if input_data != "":
                    # Run the networks only if we could generate the input data.
                    network_answers=self.run_networks(self._no_dep_networks, input_data)
                    
                    for network in self._no_dep_networks:
                        # Read each network's output, but only store
                        # it if there's anything relevant.
                        stdout = network.stdout()
                        stderr = network.stderr()
                        
                        if len(stdout) > 0:
                            network_stdout[network.name()] = stdout
                            
                        if len(stderr) > 0:
                            network_stderr[network.name()] = stderr
                else:
                    # We couldn't generate the input data for the ANN
                    # so there's no point in further processing.
                    TP.PrintLine("\tSkipping...")
                    
            answers.append({
                'bottom': layer['bottom'],
                'top': layer['top'],
                'networks': network_answers,
                'depolarisation': has_dep
            })
            
        result=NNVotes(answers)

        return result, network_stdout, network_stderr

    def run_networks (self, networks, input_data):
        '''
        Run the ANNs.
        
        Keyword arguments:
        networks   -- list of ANNs to run.
        input_data -- data to write to the input file
                      of the ANN.
        '''
        networks_running=0
        number_of_aerosol_types=14
        
        answers=[]
        
        # Start each neural network
        
        for network in networks:
            input_file=open(network.input_filename(), "w")
            input_file.write(input_data)
            input_file.close()
            
            network.start()
            networks_running += 1
        
        # Wait for them to finish
        
        # sys.stdout.write('\tWaiting for ANNs...')
        TP.Print('\tWaiting for ANNs...')
        
        stopped_networks=[]
        
        while True:
            for network in networks:
                if (network.running() == False and
                    network.name() not in stopped_networks
                    ):
                    networks_running -= 1
                    stopped_networks.append(network.name())
                    
            if self._thread.Stopped():
                raise RuntimeError('Work aborted')
                
            if networks_running < 1:
                # sys.stdout.write(' ')
                TP.Print(' ')
                break
                
            time.sleep(0.5) # Wait 0.5s before polling again
            # sys.stdout.write('.')
            TP.Print('.')
            
        TP.PrintLine("Done. Collecting results.")
            
        for network in networks:
            answers.append({
                "name": network.name(),
                "result": network.result(self._min_probability),
            })
            
        return answers
        
    def stop_networks(self, networks=[]):
        '''
        Stop the running ANNs.
        
        Keyword arguments:
        networks -- list of currently running ANNs.
        '''
        stdout={}
        stderr={}
        
        for network in networks:
            # Stop the network
            network.stop()
            
            # Read its output (if any)
            o = network.stdout()
            if len(o) > 0:
                stdout[network.name()] = o
            
            # Read its errors (if any)
            e = network.stderr()
            if len(e) > 0:
                stderr[network.name()] = e
            
        return stdout, stderr
        
class Election:
    '''
    Choose final result based on multiple
    ANN answers.
    '''
    def __init__(self, votes=None, min_ratio=.25):
        '''
        Create the Election object.
        
        Keyword arguments:
        votes     -- list of NNVotes objects.
        min_ratio -- the minimum ratio of confident
                     answers a network has to give.
        '''
        self._votes=votes
        self._min_ratio=min_ratio
        
    def results(self):
        '''
        Get the results of the election.
        '''
        result=[]
        
        for layer in self._votes.per_layer():
            l_typing=(len(layer['L_answers']) > 0)
            h_typing=(len(layer['H_answers']) > 0)
            
            L_result=('N/A', {}, False)
            H_result=('N/A', {}, False)
            
            if (l_typing == False and h_typing == False):
                # We didn't run the networks for this layer
                # but we shouldn't print any error message
                # as this has been taken care of inside natali.py
                pass
            elif l_typing == False:
                # H-resolution typing is possible
                H_result=self.compute_result(layer['H_answers'])
                
                if H_result[2] == False:
                    # Couldn't determine the aerosol type
                    H_result[1]['Comments'] = "No typing possible: " + H_result[1]['Comments']
            elif h_typing == False:
                # L-resolution typing is possible
                L_result=self.compute_result(layer['L_answers'])
                
                if L_result[2] == False:
                    # Couldn't determine the aerosol type
                    L_result[1]['Comments'] = "No typing possible: " + L_result[1]['Comments']
            else:
                # Typing is possible for both H-resolution and
                # L-resolution
                H_result=self.compute_result(layer['H_answers'])
                L_result=self.compute_result(layer['L_answers'])
                
                if (H_result[2] == False and L_result[2] == False):
                    # The typing has failed for both the aerosol type
                    # and the predominant component.
                    L_result[1]['Comments'] = "No typing possible: no ANNs passed the minimum agreement criteria. "
                    H_result[1]['Comments'] = ''
                elif L_result[2] == False:
                    # Typing has failed for the predominant component
                    L_result[1]['Comments'] = "No typing possible for L-resolution: " + L_result[1]['Comments']
                elif H_result[2] == False:
                    # Typing has failed for the aerosol type
                    H_result[1]['Comments'] = "No typing possible for H-resolution: " + H_result[1]['Comments']
    
            result.append({
                'bottom': layer['bottom'],
                'top': layer['top'],
                'L_result': L_result,
                'H_result': H_result
            })
            
        return result
        
        
    def compute_result(self, layer):
        '''
        Compute the election result for a given layer.
        '''
        if len(layer) == 0:
            # Networks didn't have answers for this layer.
            # Respond with "N/A"
            return "N/A", {'Comments': ''}, False
            
        possible_results=[
            'Coastal/CC', 'CoastalPolluted/CC', 'Continental', 'ContinentalDust',
            'ContinentalPolluted', 'ContinentalSmoke', 'Dust', 'DustPolluted',
            'Marine/CC', 'MineralMixtures/Volcanic', 'MixedDust/CC', 'MixedSmoke/CC', 'Smoke',
            'MineralMixtures/Volcanic', 'Unknown'
        ]
        
        votes={}
        output={'Comments': ""}
        for i in range(0, len(possible_results)):
            votes[possible_results[i]]=[]
            
        answer='Unknown'
        
        for network, result in layer:
            aerosol_vote=('Unknown', 0, .0)
            
            for type, data in result.answers().items():
                
                if (data["trust_level"] > aerosol_vote[2] and
                    data["confident_answers"] > self._min_ratio*result.count()):
                    
                    aerosol_vote=(type, data["confident_answers"], data["trust_level"])
            
            if aerosol_vote[1] > 0:
                votes[aerosol_vote[0]].append({
                    'network': network,
                    'confident_answers': aerosol_vote[1],
                    'confidence': result.answers()[aerosol_vote[0]]["confidence"],
                    'trust_level': aerosol_vote[2],
                })
                
                output[network + "_Answer"]="%s" % (aerosol_vote[0])
                output[network + "_Confidence"]="%.2f" % (result.answers()[aerosol_vote[0]]["confidence"])
                output[network + "_Agreements"]="%d" % aerosol_vote[1]
                
            else:
                output[network + "_Answer"]="Unknown"
                
        max_votes = 0
        for v in votes.values():
            max_votes=max(max_votes, len(v))
            
        if max_votes == 0:
            # Decide which type of ANNs we ran.
            # positive value -> L-type
            # negative value -> H-type
            l_vs_h=0
            for network, result in layer:
                if network.endswith('L'):
                    l_vs_h += 1
                elif network.endswith('H'):
                    l_vs_h -= 1
                    
            if l_vs_h > 0:
                # L-type ANNs
                output['Comments'] += "No L-type ANN passed the minimum agreement criteria. "
            elif l_vs_h < 0:
                # H-type ANNs
                output['Comments'] += "No H-type ANN passed the minimum agreement criteria. "
                
            return answer, output, False
        
        max_trust_votes=0
        max_trust_networks={}

        # Compile network votes
        
        max_trust_level = 0
        
        for aerosol, networks in votes.items():
            if aerosol == "Unknown":
                # Ignore networks which haven't
                continue
                
            for network in networks:
                trust = network["confident_answers"] * network["confidence"]
                
                if trust > max_trust_level:
                    answer = aerosol
                    max_trust_level = trust
    
        return answer, output, True
        
def scramble(layers, samples=50):
    '''
    Generates multiple values inside the
    [mean-uncertainty; mean+uncertainty] interval
    for each of the intensive parameters, and
    mixes them up to obtain multiple parameter value
    sets (hence 'scramble').
    
    Keyword arguments:
    layers  -- array of layers containing the intensive
               parameters
    samples -- number of values to be generated
               inside the [mean-uncertainty; mean+uncertainty]
               interval
    '''
    scrambled = []
    
    for layer in layers:
        
        scrambled_parameters=[]
        
        parameters=[
            'ae_355_532', 'ci_355_532', 'ci_532_1064',
            'cr_355_532', 'cr_532_1064', 'lr_355', 'lr_532'
        ]
        
        # Check that we have computed all the necessary intensive parameters:
        
        if (layer["ae_355_532"]["range_flag"] == True or layer["ae_355_532"]["nan_flag"] == True or
            layer["ci_355_532"]["range_flag"] == True or layer["ci_355_532"]["nan_flag"] == True or
            layer["ci_532_1064"]["range_flag"] == True or layer["ci_532_1064"]["nan_flag"] == True or
            layer["cr_355_532"]["range_flag"] == True or layer["cr_355_532"]["nan_flag"] == True or
            layer["cr_532_1064"]["range_flag"] == True or layer["cr_532_1064"]["nan_flag"] == True or
            layer["lr_355"]["range_flag"] == True or layer["lr_355"]["nan_flag"] == True or
            layer["lr_532"]["range_flag"] == True or layer["lr_532"]["nan_flag"] == True
            ):
            
            scrambled.append({
                "bottom": layer["bottom"],
                "top": layer["top"],
                "parameters": [dict(zip(parameters, [np.nan]*len(parameters)))]
            })
            
            continue
        
        # Generate all possible values within the error bar:

        distributed=dict(zip(
            parameters,
            [np.linspace(
                layer[param]["average"]-layer[param]["error"],
                layer[param]["average"]+layer[param]["error"],
                samples
            
            ) for param in parameters]
        ))
        
        if "dep" in layer.keys():
            if layer["dep"]["range_flag"] != True or layer["dep"]["nan_flag"] != True:
                distributed["dep"]=np.linspace(
                    layer["dep"]["average"]-layer["dep"]["error"],
                    layer["dep"]["average"]+layer["dep"]["error"],
                    samples
                )
        
        parameters=[]
        
        for name, series in distributed.items():
            for value in series:
                for i in range (0, samples):
                    param_set = {name: value}
                    for pname in distributed.keys():
                        if pname is not name:
                            param_set[pname] = distributed[pname][i]
                            
                    parameters.append(param_set)
        
        # And add these combinations to the current layer:
        scrambled.append({
            "bottom": layer["bottom"],
            "top": layer["top"],
            "parameters": parameters
        })
        
    return scrambled