from collections import OrderedDict
import os
import sys
import wx
from wx.adv import BitmapComboBox

import nt_globals as NtGlobals
import nt_data as NtData
import nt_output as NtOutput
from nt_output import TextPrinter as TP

import matplotlib
import numpy as np
from numpy import arange, sin, pi

matplotlib.use('WXAgg')
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as FigureCanvas
from matplotlib.figure import Figure
from matplotlib.ticker import FuncFormatter
import matplotlib.patches as mpatches

import json

class MainWindow(wx.Frame):
    '''
    Use to display the main window
    of our application.
    '''
    
    ID_TOOL_FOLDER    = 101
    ID_TOOL_DATASETS  = 102
    ID_TOOL_STARTSTOP = 103
    ID_TOOL_SETTINGS  = 104
    ID_TOOL_VIEWLOG   = 105
    
    ID_FIGURE_1       = 201
    ID_FIGURE_2       = 202
    ID_FIGURE_3       = 203
    ID_FIGURE_4       = 204
    ID_FIGURE_L       = 205
    
    def __init__(self, title):
        '''
        Construct a MainWindow object.
        
        Keyword arguments:
        title -- the window title.
        '''
        # Initialize members.
        
        self._running = False
        self._dataFolder = ''
        self._datasets = OrderedDict()
        self._progress = 0
        self.worker = None
        self._processedDatasets = OrderedDict()
        self._settings = {
            'ANN_Folder': NtGlobals.ANN_FOLDER,
            'filterWindow': NtGlobals.FILTER_WINDOW,
            'minLayerDepth': NtGlobals.MIN_LAYER_DEPTH,
            'finesse': NtGlobals.FINESSE,
            'minAcceptedConfidence': NtGlobals.MIN_ACCEPTED_CONFIDENCE,
            'minAgreementPercentage': NtGlobals.MIN_AGREEMENT_PERCENTAGE,
            'altitudeScaleLow': NtGlobals.ALTITUDE_SCALE_LOW,
            'altitudeScaleHigh': NtGlobals.ALTITUDE_SCALE_HIGH,
            'highResolution': True
        }
        
        self._activeDataset = ''
        self.folderDialogOpened = False
        self.datasetDialogOpened = False
        
        # Construct the wx.Frame
        
        super(MainWindow, self).__init__(None, title=title)
        
        # Set the initial size, do not allow smaller sizes
        # and center the window on the screen.
        
        self.SetSize(wx.Size(1050, 750))
        self.SetMinSize(self.GetSize())
        
        # Build the statusbar.
        
        self.statusBar = self.CreateStatusBar()
        self.statusBar.SetStatusText('Idle. Please choose data folder.')
        
        self.Center()
        
        # Create the toolbar
        
        self.toolbar = self.CreateToolBar(wx.TB_HORZ_TEXT)
        
        self.folderTool = self.toolbar.AddLabelTool(self.ID_TOOL_FOLDER, 'Choose data  folder', wx.Bitmap('icons/folder.png'))
        self.Bind(wx.EVT_TOOL, self.OnSelectFolderBtnClicked, id=self.ID_TOOL_FOLDER)
        
        self.datasetsTool = self.toolbar.AddLabelTool(self.ID_TOOL_DATASETS, 'Choose datasets', wx.Bitmap('icons/page.png'))
        self.Bind(wx.EVT_TOOL, self.OnSelectDatasetBtnClicked, id=self.ID_TOOL_DATASETS)
        
        self.startStopTool = self.toolbar.AddLabelTool(self.ID_TOOL_STARTSTOP, 'Start', wx.Bitmap('icons/start.png'))
        self.Bind(wx.EVT_TOOL, self.OnStartStopToolClicked, id=self.ID_TOOL_STARTSTOP)
        
        self.toolbar.AddSeparator()
        
        self.settingsTool = self.toolbar.AddLabelTool(self.ID_TOOL_SETTINGS, 'Settings', wx.Bitmap('icons/settings.png'))
        self.Bind(wx.EVT_TOOL, self.OnSettingsBtnClicked, id=self.ID_TOOL_SETTINGS)
        
        # self.viewLogTool = self.toolbar.AddLabelTool(self.ID_TOOL_VIEWLOG, 'View log', wx.Bitmap('icons/comment.png'))
        
        self.toolbar.Realize()
        
        # Build the main layout.
        
        self.panel = wx.Panel(self, style=wx.EXPAND)
        
        MasterSizer = wx.BoxSizer(wx.HORIZONTAL) # Wrapper
        TopLeftVBoxSizer = wx.BoxSizer(wx.VERTICAL) # Top left box sizer
        SelectDatasetHBoxSizer = wx.BoxSizer(wx.HORIZONTAL) # "Select dataset" dropdown box sizer
        VBoxSizer = wx.BoxSizer(wx.VERTICAL) # Main vertical box sizer
        HBoxSizer = wx.BoxSizer(wx.HORIZONTAL) # Top horizontal box sizer
        HBox2Sizer = wx.BoxSizer(wx.HORIZONTAL) # Lower horizontal box sizer
        
        # Add widgets to layout.
        
        # Start with the top left quarter.
        
        self.logViewer = wx.TextCtrl(self.panel, style=wx.TE_MULTILINE|wx.TE_READONLY|wx.TE_RICH)
        self.logViewer.SetDefaultStyle(wx.TextAttr(colText=wx.BLACK, colBack=wx.WHITE, font=wx.Font(10, wx.FONTFAMILY_TELETYPE, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL)))
        
        self.datasetComboBox = wx.adv.BitmapComboBox(self.panel, choices=['No dataset selected'], style=wx.CB_READONLY)
        self.datasetComboBox.Disable()
        self.datasetComboBox.Select(0)
        self.datasetComboBox.Bind(wx.EVT_COMBOBOX, self.OnDatasetComboBoxChanged)
        
        self._errorIcon = wx.Bitmap('icons/error.png')
        self._successIcon = wx.Bitmap('icons/success.png')
        self._warningIcon = wx.Bitmap('icons/warning.png')
        
        self.datasetComboBoxLabel = wx.StaticText(self.panel, wx.ID_ANY, 'View data for:')
        
        SelectDatasetHBoxSizer.Add(self.datasetComboBoxLabel, proportion=0, flag=wx.ALL, border=5)
        SelectDatasetHBoxSizer.Add(self.datasetComboBox, proportion=1, flag=wx.EXPAND|wx.ALL, border=5)
        
        TopLeftVBoxSizer.Add(self.logViewer, proportion=1, flag=wx.EXPAND|wx.ALL, border=5)
        TopLeftVBoxSizer.Add(SelectDatasetHBoxSizer, proportion=0, flag=wx.EXPAND|wx.ALL, border=5)
        
        # Continue with the timeseries plot.
        
        self.largeFigure = NtFigure(self.panel, self.ID_FIGURE_L)
        self.largeFigure.Bind(wx.EVT_RIGHT_DOWN, self.OnFigureRightClick)
        
        # Add the widgets to the top horizontal box sizer.
        
        HBoxSizer.Add(TopLeftVBoxSizer, proportion=1, flag=wx.EXPAND|wx.ALL, border=5)
        HBoxSizer.Add(self.largeFigure, proportion=1, flag=wx.EXPAND|wx.ALL, border=5)
        
        # Add the four bottom figures.
        
        self.figure1 = NtFigure(self.panel, self.ID_FIGURE_1)
        self.figure1.Bind(wx.EVT_RIGHT_DOWN, self.OnFigureRightClick)
        
        self.figure2 = NtFigure(self.panel, self.ID_FIGURE_2)
        self.figure2.Bind(wx.EVT_RIGHT_DOWN, self.OnFigureRightClick)
        
        self.figure3 = NtFigure(self.panel, self.ID_FIGURE_3)
        self.figure3.Bind(wx.EVT_RIGHT_DOWN, self.OnFigureRightClick)
        
        self.figure4 = NtFigure(self.panel, self.ID_FIGURE_4)
        self.figure4.Bind(wx.EVT_RIGHT_DOWN, self.OnFigureRightClick)
        
        self.ClearFigure1()
        self.ClearFigure2()
        self.ClearFigure3()
        self.ClearFigure4()
            
        self.largeFigure.Clear(settings={
            'xaxis': None,
            'yaxis': {
                'limits': (self._settings['altitudeScaleLow'], self._settings['altitudeScaleHigh']),
                'label': 'Altitude (km)'
            },
            'aerosolLegend': True
        })
        
        HBox2Sizer.Add(self.figure1, proportion=1, flag=wx.EXPAND|wx.ALL, border=5)
        HBox2Sizer.Add(self.figure2, proportion=1, flag=wx.EXPAND|wx.ALL, border=5)
        HBox2Sizer.Add(self.figure3, proportion=1, flag=wx.EXPAND|wx.ALL, border=5)
        HBox2Sizer.Add(self.figure4, proportion=1, flag=wx.EXPAND|wx.ALL, border=5)
        
        # Finally, glue everything together!
        
        VBoxSizer.Add(HBoxSizer, proportion=1, flag=wx.EXPAND)
        VBoxSizer.Add(HBox2Sizer, proportion=1, flag=wx.EXPAND)
        MasterSizer.Add(VBoxSizer, proportion=1, flag=wx.ALL|wx.EXPAND, border=5)
        
        self.panel.SetSizer(MasterSizer)
        
        # Set up initial status of widgets.
        
        self.toolbar.EnableTool(self.ID_TOOL_FOLDER, True)
        self.toolbar.EnableTool(self.ID_TOOL_DATASETS, False)
        self.toolbar.EnableTool(self.ID_TOOL_STARTSTOP, False)
        self.toolbar.EnableTool(self.ID_TOOL_SETTINGS, True)
        
        # Register event handlers.
        
        self.Bind(NtData.EVT_WORK_STARTED, self.OnWorkStarted)
        self.Bind(NtData.EVT_WORK_ABORTED, self.OnWorkAborted)
        self.Bind(NtData.EVT_WORK_FINISHED, self.OnWorkFinished)
        
        self.Bind(NtData.EVT_DATASET_STARTED, self.OnDatasetStarted)
        self.Bind(NtData.EVT_DATASET_FINISHED, self.OnDatasetFinished)
        
        self.Bind(NtData.EVT_DATASET_ERROR, self.OnDatasetError)
        self.Bind(NtData.EVT_DATASET_WARNING, self.OnDatasetWarning)
        
        self.Bind(NtData.EVT_EXT_PARAMS_READ, self.OnExtParamsRead)
        self.Bind(NtData.EVT_INT_PARAMS_COMPUTED, self.OnIntParamsComputed)
        
        self.Bind(NtData.EVT_LAYERS_COMPUTED, self.OnLayersComputed)
        self.Bind(NtData.EVT_AEROSOLS_COMPUTED, self.OnAerosolsComputed)
        
        self.Bind(NtOutput.EVT_PRINT_TEXT, self.OnPrintText)
        
    def OnPrintText(self, e):        
        lastLineNumber = self.logViewer.PositionToXY(self.logViewer.LastPosition)[1]
        
        test = self.logViewer.HitTestPos((0, 5))[1] # 5 pixels from top
        firstVisibleLine = self.logViewer.PositionToXY(test)[1]
        
        test = self.logViewer.HitTestPos((0, self.logViewer.GetSizeTuple()[1]-5))[1] # 5 pixels from bottom
        
        lastVisibleLine = self.logViewer.PositionToXY(test)[1]
        
        if lastLineNumber <= lastVisibleLine:
            # At the bottom
            self.logViewer.AppendText(e.data)
        else:
            self.logViewer.Freeze()
            
            firstPosition = self.logViewer.HitTestPos((0, 5))[1]
            
            self.logViewer.ChangeValue(self.logViewer.GetValue() + e.data)
            self.logViewer.SetStyle(0, self.logViewer.GetLastPosition(), self.logViewer.GetDefaultStyle())
            
            self.logViewer.ShowPosition(firstPosition)
            
            self.logViewer.Thaw()
        
    def OnSelectFolderBtnClicked(self, e):
        '''
        "Select folder" toolbar tool event handler.
        
        Keyword arguments:
        e -- the event.
        '''
        if self.folderDialogOpened:
            return
            
        dirDialog = wx.DirDialog(
            self,
            "Choose data folder:",
            style = wx.DD_DIR_MUST_EXIST
        )
        
        self.folderDialogOpened = True
        
        if dirDialog.ShowModal() == wx.ID_OK:
            # Selected folder changed!
            self.OnDataFolderChanged(dirDialog.GetPath())
            
        dirDialog.Destroy()
        
        self.folderDialogOpened = False
            
    def OnDataFolderChanged(self, newfolder):
        '''
        Called anytime when a new data folder is selected
        by the user.
        
        Keyword arguments:
        newfolder -- the full path of the new folder.
        '''
        if self._dataFolder == newfolder:
            # Don't change anything if the user selected the
            # same folder as before.
            return
            
        self._dataFolder = newfolder
        
        measurements = NtGlobals.ReadFolder ( self._dataFolder )
            
        for measurement in measurements:
            measurement[ "selected" ] = True
        
        if (len(measurements) > 0):
            self.toolbar.EnableTool(self.ID_TOOL_DATASETS, True)
            self.toolbar.Realize()
            self.Layout()
            
            self.OnDatasetsChanged(measurements)
        else:
            self.OnDatasetsChanged(OrderedDict())
    
    def OnSelectDatasetBtnClicked(self, e):
        '''
        "Select dataset" toolbar tool event handler.
        
        Keyword arguments:
        e -- the event.
        '''
        
        if self.datasetDialogOpened:
            return
        
        dialog = DatasetsDialog(self, title="Choose dataset", datasets=self._datasets)
        
        self.datasetDialogOpened = True
        
        if dialog.ShowModal() == wx.ID_OK:
            self.OnDatasetsChanged(dialog.GetDatasets())
            
        self.UpdateStatusBar()
        
        dialog.Destroy()
        
        self.datasetDialogOpened = False
        
    def OnDatasetsChanged(self, datasets):
        '''
        Called anytime when the list of selected
        datasets is changed.
        
        Keyword arguments:
        datasets -- the new list of datasets.
        '''
        self._datasets = datasets
        
        if self._dataFolder != '':
            self.datasetsTool.SetLabel("Choose datasets (%d/%d)" % (self.SelectedDatasetsCount(), self.TotalDatasetsCount()))
        else:
            self.datasetsTool.SetLabel("Choose datasets")
            
        self.UpdateStatusBar()
        
        if self.SelectedDatasetsCount() > 0:
            self.toolbar.EnableTool(self.ID_TOOL_STARTSTOP, True)
        else:
            self.toolbar.EnableTool(self.ID_TOOL_STARTSTOP, False)
            
        self.toolbar.Realize()
        self.Layout()
        
    def TotalDatasetsCount(self):
        '''
        Get the total number of datasets
        inside the data folder.
        '''
        return len(self._datasets)
        
    def SelectedDatasetsCount(self):
        '''
        Get the total number of
        selected datasets.
        '''
        c = 0
        
        for measurement in self._datasets:
            if measurement["selected"] == True:
                c += 1
                
        return c
        
    def GetSelectedDatasets(self):
        '''
        Get a list of seected datasets.
        '''
        d = []
        
        for measurement in self._datasets:
            if measurement["selected"] == True:
                d.append(measurement)
                
        return d
    
    def OnStartStopToolClicked(self, e):
        '''
        "Start/Stop" toolbar tool event handler.
        
        Keyword arguments:
        e -- the event.
        '''
        if self._running == False:
            self.StartWorker()
        else:
            self.StopWorker()
            
        self.UpdateStatusBar()
        
    def OnSettingsBtnClicked(self, e):
        dialog = SettingsDialog(self, title="Settings", settings=self._settings)
        
        if dialog.ShowModal() == wx.ID_OK:
            self.OnSettingsChanged(dialog.GetSettings())
            
        dialog.Destroy()
            
    def OnDatasetComboBoxChanged(self, e):
        self.OnActiveDatasetChanged(e.GetString())
        
    def OnActiveDatasetChanged(self, dataset):
        # Scroll to the relevant dataset in the log viewer.
        
        position = self.logViewer.GetValue().find(dataset)
        self._activeDataset = dataset
        
        self.largeFigure.PlotAerosolTimeseries(
            self._processedDatasets,
            ylimits=(self._settings['altitudeScaleLow'], self._settings['altitudeScaleHigh']),
            activeDataset=self._activeDataset,
            highResolution=self._settings['highResolution']
        )
        
        if position != -1:
            self.logViewer.ShowPosition(position)
        
        if 'extparams' in self._processedDatasets[dataset].keys():
            self.PlotFigure1(dataset)
        else:
            self.ClearFigure1('Not available.')
            self.ClearFigure2('Not available.')
            self.ClearFigure3('Not available.')
            
        if 'intparams' in self._processedDatasets[dataset].keys():            
            self.PlotFigure2(dataset)
            self.PlotFigure3(dataset)
        else:
            self.ClearFigure3('Not available.')
        
        if 'results' in self._processedDatasets[dataset].keys():            
            self.PlotFigure4(dataset)
        else:
            self.ClearFigure4('Not available.')
    
    def OnSettingsChanged(self, newsettings):    
        self._settings = newsettings
        
        if self._activeDataset != '':
            self.OnActiveDatasetChanged(self._activeDataset)
            self.largeFigure.PlotAerosolTimeseries(
                self._processedDatasets,
                ylimits=(self._settings['altitudeScaleLow'], self._settings['altitudeScaleHigh']),
                activeDataset=self._activeDataset,
                highResolution=self._settings['highResolution']
            )
        else:
            self.ClearFigure1()
            self.ClearFigure2()
            self.ClearFigure3()
            self.ClearFigure4()
                
            self.largeFigure.Clear(settings={
                'xaxis': None,
                'yaxis': {
                    'limits': (self._settings['altitudeScaleLow'], self._settings['altitudeScaleHigh']),
                    'label': 'Altitude (km)'
                },
                'aerosolLegend': True
            })
        
    def UpdateStatusBar(self):
        '''
        Update the status bar text to reflect the current
        state of the application.
        '''
        
        # TODO: Color the status bar if possible.
        
        status = ''
        
        if self._running == True:
            status = "Processing dataset %d of %d." % (self._progress, self.SelectedDatasetsCount())
        else:
            status = "Idle."
            
            if self._dataFolder != '':
                status += " %d of %d datasets selected from folder %s" % (self.SelectedDatasetsCount(), self.TotalDatasetsCount(), self._dataFolder)
            else:
                status += " Please choose data folder."
    
        self.statusBar.SetStatusText(status)
        
    def StartWorker(self):
        '''
        Start the processing thread for the selected datasets.
        '''        
        if self.worker is not None:
            self.StopWorker()
            
        self.worker = NtData.NtWorkerThread(
            parent=self,
            datasets=self.GetSelectedDatasets(),
            folder=self._dataFolder,
            ann_folder = self._settings['ANN_Folder'],
            filter_window = self._settings['filterWindow'],
            min_layer_depth = self._settings['minLayerDepth'],
            finesse = self._settings['finesse'],
            min_accepted_confidence = self._settings['minAcceptedConfidence'],
            min_agreement_percentage = self._settings['minAgreementPercentage']
        )
        
        self._progress = 0
        self.worker.start()
        self.toolbar.EnableTool(self.ID_TOOL_STARTSTOP, False)
        self.toolbar.Realize()
        self.Layout()
        
    def StopWorker(self):
        '''
        Stop the processing thread immediately.
        '''
        
        if self.worker == None:
            return
            
        self.worker.stop()
        self.toolbar.EnableTool(self.ID_TOOL_STARTSTOP, False)
        self.toolbar.Realize()
        self.Layout()
        
    def OnWorkStart(self):
        '''
        Called right before starting the processing thread.
        '''
        self._running = True
        
        self.startStopTool.SetLabel("Stop")
        self.startStopTool.SetNormalBitmap(wx.Bitmap('icons/stop.png'))
        
        # Disable the settings button once the processing starts,
        # and re-enable the start/stop button.
        self.toolbar.EnableTool(self.ID_TOOL_SETTINGS, False)
        self.toolbar.EnableTool(self.ID_TOOL_FOLDER, False)
        self.toolbar.EnableTool(self.ID_TOOL_DATASETS, False)
        self.toolbar.EnableTool(self.ID_TOOL_STARTSTOP, True)
        self.toolbar.Realize()
        self.Layout()
        
        # self.largeFigure.Clear(settings={'xaxis': None, 'yaxis': {'limits': (self._settings['altitudeScaleLow'], self._settings['altitudeScaleHigh']), 'label': "Altitude(km)"}})
        
        self._processedDatasets = OrderedDict([(ds["measurement"], {'has_warning': False, 'warning': '', 'has_error': False, 'error': ''}) for ds in self.GetSelectedDatasets()])
        self.largeFigure.PlotAerosolTimeseries(
            self._processedDatasets,
            ylimits=(self._settings['altitudeScaleLow'], self._settings['altitudeScaleHigh']),
            highResolution=self._settings['highResolution']
        ) # Initialize graph
        
        self.datasetComboBox.SetItems([ds["measurement"] for ds in self.GetSelectedDatasets()])
        self.datasetComboBox.Select(0)
        
        self._activeDataset = ''
        
        self.UpdateStatusBar()
        
    def OnWorkEnd(self):
        '''
        Called right after work, or right after
        stopping the processing thread.
        '''
        self._running = False
        
        self.startStopTool.SetLabel("Start")
        self.startStopTool.SetNormalBitmap(wx.Bitmap('icons/start.png'))
        
        # Re-enable the settings and start/stop button.
        self.toolbar.EnableTool(self.ID_TOOL_SETTINGS, True)
        self.toolbar.EnableTool(self.ID_TOOL_FOLDER, True)
        self.toolbar.EnableTool(self.ID_TOOL_DATASETS, True)
        self.toolbar.EnableTool(self.ID_TOOL_STARTSTOP, True)
        self.toolbar.Realize()
        self.Layout()
        
        self.largeFigure.PlotAerosolTimeseries(
            self._processedDatasets,
            ylimits=(self._settings['altitudeScaleLow'], self._settings['altitudeScaleHigh']),
            activeDataset=self._activeDataset,
            highResolution=self._settings['highResolution']
        )
        
        for name, dataset in self._processedDatasets.items():
            if dataset['has_error'] == True:
                self.datasetComboBox.SetItemBitmap(self.datasetComboBox.FindString(name), self._errorIcon)
            elif dataset['has_warning'] == True:
                self.datasetComboBox.SetItemBitmap(self.datasetComboBox.FindString(name), self._warningIcon)
            elif dataset:
                self.datasetComboBox.SetItemBitmap(self.datasetComboBox.FindString(name), self._successIcon)
                
        self.datasetComboBox.SetSelection(0)
        self.datasetComboBox.Enable()
        
        self.UpdateStatusBar()

    def OnWorkStarted(self, e):
        self.OnWorkStart()
    
    def OnWorkFinished(self, e):        
        self.OnWorkEnd()
        
    def OnWorkAborted(self, e):
        TP.Log('Processing aborted by user!')
        TP.PrintLine('Processing aborted by user!')
        self.OnWorkEnd()
        
    def OnDatasetWarning(self, e):
        self._processedDatasets[e.data['dataset']]['has_warning'] = True
        self._processedDatasets[e.data['dataset']]['warning'] += e.data['warning']
        
        TP.PrintLine('\n%s warning: %s' % (e.data['dataset'], e.data['warning']))
    
    def OnDatasetError(self, e):
        self._processedDatasets[e.data['dataset']]['has_error'] = True
        self._processedDatasets[e.data['dataset']]['error'] += e.data['error']
        
        print "\"%s\" error: %s" % (e.data['dataset'], e.data['error'])
        
    def OnExtParamsRead(self, e):
        self._processedDatasets[e.data['dataset']]['altitudes']=e.data['altitudes']
        self._processedDatasets[e.data['dataset']]['extparams']=e.data['params']
        self._processedDatasets[e.data['dataset']]['has_dep']=e.data['has_dep']
        
        self.PlotFigure1(e.data['dataset'])
        self.PlotFigure2(e.data['dataset'])
        
        self.ClearFigure3('Computing layers...')
            
        self.ClearFigure4('Computing...')
              
    def OnIntParamsComputed(self, e):
        self._processedDatasets[e.data['dataset']]['intparams']=e.data['params']
        
        self.PlotFigure3(e.data['dataset'])
        self.PlotFigure4(e.data['dataset'])
        # self.ClearFigure4('Computing...')
        
    def OnLayersComputed(self, e):
        if self._processedDatasets[e.data['dataset']]['has_dep'] == False:
            self.ClearFigure3('No linear depolarisation ratio data.')
            return
            
        bottom = 0
        
        for ilayer in e.data['ilayers']:
            for i in range(bottom, ilayer[0]-1):
                self._processedDatasets[e.data['dataset']]['extparams']['dep']['data'][i] = np.nan
                self._processedDatasets[e.data['dataset']]['extparams']['dep']['error'][i] = np.nan
                
            bottom = ilayer[1]+1
            
        for i in range (bottom, len(self._processedDatasets[e.data['dataset']]['extparams']['dep']['data'])):
            self._processedDatasets[e.data['dataset']]['extparams']['dep']['data'][i] = np.nan
            self._processedDatasets[e.data['dataset']]['extparams']['dep']['error'][i] = np.nan
        
        self.PlotFigure3(e.data['dataset'], onlydep=True)
        
    def OnAerosolsComputed(self, e):
        self._processedDatasets[e.data['dataset']]['results']=e.data['results']
        
        self.largeFigure.PlotAerosolTimeseries(
            self._processedDatasets,
            ylimits=(self._settings['altitudeScaleLow'], self._settings['altitudeScaleHigh']),
            activeDataset=self._activeDataset,
            highResolution=self._settings['highResolution']
        )
    
    def OnDatasetStarted(self, e):
        self._progress += 1
        self.UpdateStatusBar()
        
        self.ClearFigure1('Reading data...')
        self.ClearFigure2('Reading data...')
        
        self.datasetComboBox.SetSelection(self._progress - 1)
        self._activeDataset = e.data['dataset']

        self._processedDatasets[e.data['dataset']] = {'has_warning': False, 'warning': '', 'has_error': False, 'error': ''}
        
        self.largeFigure.PlotAerosolTimeseries(
            self._processedDatasets,
            ylimits=(self._settings['altitudeScaleLow'], self._settings['altitudeScaleHigh']),
            activeDataset=self._activeDataset,
            highResolution=self._settings['highResolution']
        )
    
    def OnDatasetFinished(self, e):
        if self._processedDatasets[e.data['dataset']]['has_warning'] == False and \
           self._processedDatasets[e.data['dataset']]['has_error']   == False:
           
            self.datasetComboBox.SetItemBitmap(self.datasetComboBox.FindString(e.data['dataset']), self._successIcon)

        self.datasetComboBox.SetSelection(self._progress - 1)
    
    def PlotFigure1(self, dataset):
        self.figure1.PlotMultipleHorizontal(
            ylabel="Altitude (km)",
            ydata=[x / 1000.0 for x in self._processedDatasets[dataset]['altitudes']],
            ylimits=(self._settings['altitudeScaleLow'], self._settings['altitudeScaleHigh']),
            clear=True,
            axes=[{
                    'xlimits': (0, 0.00001),
                    'color': 'Black',
                    'label': 'Backscatter Coefficient [1/(m*sr)]',
                    'format': 'scientific',
                    'plots': [
                        {'name': 'b1064', 'color': 'Red', 'xdata': self._processedDatasets[dataset]['extparams']['b1064']['data'], 'xerror': self._processedDatasets[dataset]['extparams']['b1064']['error']},
                        {'name': 'b532', 'color': 'Green', 'xdata': self._processedDatasets[dataset]['extparams']['b532']['data'], 'xerror': self._processedDatasets[dataset]['extparams']['b532']['error']},
                        {'name': 'b355', 'color': 'Blue', 'xdata': self._processedDatasets[dataset]['extparams']['b355']['data'], 'xerror': self._processedDatasets[dataset]['extparams']['b355']['error']}
                    ]
                }
            ]
        )
        
    def ClearFigure1(self, text=''):
        self.figure1.Clear(
            settings={
                'xaxis': [{
                    'limits': (0, 0.00001),
                    'label': 'Backscatter Coefficient [1/(m*sr)]',
                    'color': 'Black',
                    'format': 'scientific'
                }], 'yaxis': {
                    'limits': (self._settings['altitudeScaleLow'], self._settings['altitudeScaleHigh']),
                    'label': 'Altitude (km)'
            }})
            
        self.figure1.DisplayText(text)
    
    def PlotFigure2(self, dataset):
        self.figure2.PlotMultipleHorizontal(
            ylabel="Altitude (km)",
            ydata=[x / 1000.0 for x in self._processedDatasets[dataset]['altitudes']],
            ylimits=(self._settings['altitudeScaleLow'], self._settings['altitudeScaleHigh']),
            clear=True,
            axes=[{
                    'xlimits': (0, 0.0005),
                    'color': 'Black',
                    'label': 'Extinction Coefficient [1/m]',
                    'format': 'scientific',
                    'plots': [
                        {'name': 'e355', 'color': 'Blue', 'xdata': self._processedDatasets[dataset]['extparams']['e355']['data'], 'xerror': self._processedDatasets[dataset]['extparams']['e355']['error']},
                        {'name': 'e532', 'color': 'Green', 'xdata': self._processedDatasets[dataset]['extparams']['e532']['data'], 'xerror': self._processedDatasets[dataset]['extparams']['e532']['error']}
                    ]
                }
            ]
        )
        
    def ClearFigure2(self, text=''):
        self.figure2.Clear(
            settings={
                'xaxis': [{
                    'limits': (0, 0.001),
                    'label': 'Extinction Coefficient [1/m]',
                    'color': 'Black',
                    'format': 'scientific'
                }], 'yaxis': {
                    'limits': (self._settings['altitudeScaleLow'], self._settings['altitudeScaleHigh']),
                    'label': 'Altitude (km)'
            }})
            
        self.figure2.DisplayText(text)
        
    def PlotFigure3(self, dataset, onlydep=False):
        axes=[]
        
        if onlydep == False and 'intparams' in self._processedDatasets[dataset].keys():
            axes.append({
                'xlimits': (0, 120),
                'loc': 4,
                'label': 'Lidar Ratios [sr]',
                'plots': [
                    {'name': 'lr_355', 'color': 'Blue', 'xdata': self._processedDatasets[dataset]['intparams']['lr_355']['data'], 'xerror': self._processedDatasets[dataset]['intparams']['lr_355']['error']},
                    {'name': 'lr_532', 'color': 'Green', 'xdata': self._processedDatasets[dataset]['intparams']['lr_532']['data'], 'xerror': self._processedDatasets[dataset]['intparams']['lr_532']['error']}
                ]
            })
        
        if self._processedDatasets[dataset]['has_dep'] == True:
            axes.append({
                'xlimits': (0, 40),
                'loc': 3,
                'label': 'Linear Depolarisation Ratio [%]',
                'color': 'Black',
                'plots': [
                    {'name': 'dep', 'color': 'Black', 'xdata': self._processedDatasets[dataset]['extparams']['dep']['data'], 'xerror': self._processedDatasets[dataset]['extparams']['dep']['error']}
                ]
            })
            
        self.figure3.PlotMultipleHorizontal(
            ylabel="Altitude (km)",
            ydata=[x / 1000.0 for x in self._processedDatasets[dataset]['altitudes']],
            ylimits=(self._settings['altitudeScaleLow'], self._settings['altitudeScaleHigh']),
            clear=True,
            axes=axes
        )
        
    def ClearFigure3(self, text=''):
        self.figure3.Clear(
            settings={
                    'xaxis': [{
                        'limits': (0, 120),
                        'label': 'Lidar Ratios [sr]',
                        'loc': 4
                }, {
                    'limits': (0, 40),
                    'label': 'Linear Depolarisation Ratio [%]',
                    'color': 'Black',
                    'loc': 3
                }], 'yaxis': {
                    'limits': (self._settings['altitudeScaleLow'], self._settings['altitudeScaleHigh']),
                    'label': 'Altitude (km)'
            }})
            
        self.figure3.DisplayText(text)
        
    def PlotFigure4(self, dataset):        
        self.figure4.PlotMultipleHorizontal(
            ylabel="Altitude (km)",
            ydata=[x / 1000.0 for x in self._processedDatasets[dataset]['altitudes']],
            ylimits=(self._settings['altitudeScaleLow'], self._settings['altitudeScaleHigh']),
            clear=True,
            axes=[{
                'xlimits': (-1, 3),
                'color': 'Black',
                'loc': 4,
                'label': 'Angstrom Exponent,\nColor Indexes',
                'plots': [
                    {'name': 'ae_355_532', 'color': 'Blue', 'xdata': self._processedDatasets[dataset]['intparams']['ae_355_532']['data'], 'xerror': self._processedDatasets[dataset]['intparams']['ae_355_532']['error']},
                    {'name': 'ci_355_532', 'color': 'Green', 'xdata': self._processedDatasets[dataset]['intparams']['ci_355_532']['data'], 'xerror': self._processedDatasets[dataset]['intparams']['ci_355_532']['error']},
                    {'name': 'ci_532_1064', 'color': 'Red', 'xdata': self._processedDatasets[dataset]['intparams']['ci_532_1064']['data'], 'xerror': self._processedDatasets[dataset]['intparams']['ci_532_1064']['error']},
                ]
            }]
        )
    
    def ClearFigure4(self, text=''):
        self.figure4.Clear(
            settings={
                'xaxis': {
                    'limits': (-1, 3),
                    'label': 'Angstrom Exponent,\nColor Indexes',
                    'loc': 4
                }, 'yaxis': {
                    'limits': (self._settings['altitudeScaleLow'], self._settings['altitudeScaleHigh']),
                    'label': 'Altitude (km)'
            }})
            
        self.figure4.DisplayText(text)

    def OnFigureRightClick(self, e):
        position = e.GetEventObject().GetPosition()
        position += e.GetPosition()
        
        # if e.GetEventObject() == self.largeFigure:
            # Do this to avoid hatching in final image:
        #    self.largeFigure.PlotAerosolTimeseries(self._processedDatasets, ylimits=(self._settings['altitudeScaleLow'], self._settings['altitudeScaleHigh']))
        
        self.PopupMenu(NtFigurePopupMenu(e.GetEventObject()), position)
        
        # self.largeFigure.PlotAerosolTimeseries(self._processedDatasets, ylimits=(self._settings['altitudeScaleLow'], self._settings['altitudeScaleHigh']), activeDataset=self._activeDataset)
        
class DatasetsDialog(wx.Dialog):
    '''
    Modal dialog used to select certain datasets
    from a given folder.
    '''
    def __init__(self, parent, title='Dialog', datasets=OrderedDict()):
        '''
        Construct a DatasetsDialog object.
        '''
        super(DatasetsDialog, self).__init__(None, title=title)
        
        self.SetSize((-1, 400))
        
        panel = wx.Panel(self)
        panelVBox=wx.BoxSizer(wx.VERTICAL)
        
        TopHBox=wx.BoxSizer(wx.HORIZONTAL)
        
        self.selectAllBtn = wx.Button(panel, label="Select all")
        self.selectAllBtn.Bind(wx.EVT_BUTTON, self.OnSelectAllBtnClicked)
        self.selectNoneBtn = wx.Button(panel, label="Select none")
        self.selectNoneBtn.Bind(wx.EVT_BUTTON, self.OnSelectNoneBtnClicked)
        
        TopHBox.Add(self.selectAllBtn, proportion=0, flag=wx.RIGHT, border=5)
        TopHBox.Add(self.selectNoneBtn, proportion=0, flag=wx.LEFT, border=5)
        
        self.list = wx.ListBox(panel, style=wx.LB_EXTENDED)
        
        panelVBox.Add(TopHBox, proportion=0, flag=wx.ALL, border=10)
        panelVBox.Add(self.list, proportion=1, flag=wx.EXPAND)
        
        panel.SetSizer(panelVBox)
        
        BottomHBox=wx.BoxSizer(wx.HORIZONTAL)
        
        self.applyBtn = wx.Button(self, label="OK")
        self.applyBtn.Bind(wx.EVT_BUTTON, self.OnApplyBtnClicked)
        self.cancelBtn = wx.Button(self, label="Cancel")
        self.cancelBtn.Bind(wx.EVT_BUTTON, self.OnCancelBtnClicked)
        
        BottomHBox.Add(self.applyBtn, proportion=0, flag=wx.RIGHT, border=5)
        BottomHBox.Add(self.cancelBtn, proportion=0, flag=wx.LEFT, border=5)
        
        VBox=wx.BoxSizer(wx.VERTICAL)
        VBox.Add(panel, proportion=1, flag=wx.EXPAND|wx.BOTTOM, border=10)
        VBox.Add(BottomHBox, flag=wx.CENTER|wx.ALL, border=10)
        
        self.SetSizer(VBox)
        
        self.PopulateListBox(datasets)
        self.list.EnsureVisible(0)
        
        self._initialDatasets = datasets
        self._selectedDatasets = OrderedDict()
        
    def OnSelectAllBtnClicked(self, e):
        '''
        "Select all" button event handler.
        
        Keyword arguments:
        e -- the event.
        '''
        for i in range(self.list.GetCount()):
            self.list.SetSelection(i)
            
    def OnSelectNoneBtnClicked(self, e):
        '''
        "Select none" button event handler.
        
        Keyword arguments:
        e -- the event.
        '''
        for i in range(self.list.GetCount()):
            self.list.Deselect(i)
        
    def OnApplyBtnClicked(self, e):
        '''
        "Apply" button event handler.
        
        Keyword arguments:
        e -- the event.
        '''
        names=self.list.GetItems()
        self._selectedDatasets = self._initialDatasets
        
        selectedNames = []
        
        
        for i in range(self.list.GetCount()):
            checked=self.list.IsSelected(i)
            name = names[i]
            
            if checked:
                selectedNames.append ( name )
            
        for sds in self._selectedDatasets:
            if self.DisplayName ( sds ) in selectedNames:
                sds["selected"] = True
            else:
                sds["selected"] = False
        
        self.EndModal(wx.ID_OK)
        
    def OnCancelBtnClicked(self, e):
        '''
        "Cancel" button event handler.
        
        Keyword arguments:
        e -- the event.
        '''
        self._selectedDatasets = self._initialDatasets
        self.Destroy()
        
    def PopulateListBox(self, datasets):
        '''
        Populate the selection listbox with the datasets.
        
        Keyword arguments:
        datasets -- list of datasets.
        '''
        self.list.InsertItems([self.DisplayName ( ds ) for ds in datasets], 0)
        for ds in datasets:
            if ds[ "selected" ] == True:
                self.list.SetStringSelection(self.DisplayName ( ds ), True)
                
    def GetDatasets(self):
        return self._selectedDatasets
        
    def DisplayName ( self, dataset ):
        return "%s (%s)" % ( dataset["measurement"], dataset["format_display"] )
        
class SettingsDialog(wx.Dialog):
    '''
    Modal dialog that allows the user to
    change the application settings.
    '''
    def __init__(self, parent, settings, title='Dialog'):
        super(SettingsDialog, self).__init__(None, title=title)
        
        self.SetSize((400, 350))
        
        # Set default values for parameters:
        
        self._ANN_Folder = os.path.abspath(settings['ANN_Folder'])
        self._filterWindow = settings['filterWindow']
        self._minLayerDepth = settings['minLayerDepth']
        self._finesse = settings['finesse']
        self._minAcceptedConfidence = int(settings['minAcceptedConfidence']*100)
        self._minAgreementPercentage = int(settings['minAgreementPercentage']*100)
        self._altitudeScaleLow = int(settings['altitudeScaleLow'])
        self._altitudeScaleHigh = int(settings['altitudeScaleHigh'])
        self._highResolution = bool (settings['highResolution'])
        
        # Build the layout:
        
        self.panel = wx.Panel(self)
        panelFSizer=wx.FlexGridSizer(rows=8, cols=2, vgap=5, hgap=5)
        
        self._altitudeScaleLabel = wx.StaticText(self.panel, wx.ID_ANY, 'Altitude scale (km):', style=wx.ALIGN_RIGHT)
        
        self._altitudeScaleLowSpinCtrl = wx.SpinCtrl(self.panel, value=str(self._altitudeScaleLow), size=(50, -1))
        self._altitudeScaleLowSpinCtrl.SetRange(0, self._altitudeScaleHigh-1)
        self._altitudeScaleLowSpinCtrl.Bind(wx.EVT_SPINCTRL, self.OnNumberValueChanged)
        
        self._altitudeScaleDash = wx.StaticText(self.panel, wx.ID_ANY, '-')
        
        self._altitudeScaleHighSpinCtrl = wx.SpinCtrl(self.panel, value=str(self._altitudeScaleHigh), size=(30, -1))
        self._altitudeScaleHighSpinCtrl.SetRange(self._altitudeScaleLow+1, 15)
        self._altitudeScaleHighSpinCtrl.Bind(wx.EVT_SPINCTRL, self.OnNumberValueChanged)
        
        self._altitudeScaleSpinCtrlHBoxSizer = wx.BoxSizer(wx.HORIZONTAL)
        self._altitudeScaleSpinCtrlHBoxSizer.Add(self._altitudeScaleLowSpinCtrl, proportion=1, flag=wx.EXPAND)
        self._altitudeScaleSpinCtrlHBoxSizer.Add(self._altitudeScaleDash, proportion=0, flag=wx.ALL, border=5)
        self._altitudeScaleSpinCtrlHBoxSizer.Add(self._altitudeScaleHighSpinCtrl, proportion=1, flag=wx.EXPAND)
        
        self.ANNFolderButton = wx.Button(self.panel, wx.ID_ANY, 'Choose ANN Folder')
        self.ANNFolderLabel = wx.StaticText(self.panel, wx.ID_ANY, self._ANN_Folder, style=wx.ST_ELLIPSIZE_MIDDLE)
        self.ANNFolderLabel.SetMaxSize((350, -1))
        self.ANNFolderLabel.SetSize((350, -1))
        self.ANNFolderButton.Bind(wx.EVT_BUTTON, self.OnANNFolderBtnClicked)
        
        self._filterWindowLabel = wx.StaticText(self.panel, wx.ID_ANY, 'Filter window:', style=wx.ALIGN_RIGHT)
        self._filterWindowSpinCtrl = wx.SpinCtrl(self.panel, value=str(self._filterWindow))
        self._filterWindowSpinCtrl.SetRange(300, 1000)
        self._filterWindowSpinCtrl.Bind(wx.EVT_SPINCTRL, self.OnNumberValueChanged)
        
        self._minLayerDepthLabel = wx.StaticText(self.panel, wx.ID_ANY, 'Minimum layer depth (m):', style=wx.ALIGN_RIGHT)
        self._minLayerDepthSpinCtrl = wx.SpinCtrl(self.panel, value=str(self._minLayerDepth))
        self._minLayerDepthSpinCtrl.SetRange(200, 1000)
        self._minLayerDepthSpinCtrl.Bind(wx.EVT_SPINCTRL, self.OnNumberValueChanged)
        
        self._finesseLabel = wx.StaticText(self.panel, wx.ID_ANY, 'Finesse:', style=wx.ALIGN_RIGHT)
        self._finesseSpinCtrl = wx.SpinCtrl(self.panel, value=str(self._finesse))
        self._finesseSpinCtrl.SetRange(10, 70)
        self._finesseSpinCtrl.Bind(wx.EVT_SPINCTRL, self.OnNumberValueChanged)
        
        self._minAcceptedConfidenceLabel = wx.StaticText(self.panel, wx.ID_ANY, 'Minimum accepted confidence (%):', style=wx.ALIGN_RIGHT)
        self._minAcceptedConfidenceSpinCtrl = wx.SpinCtrl(self.panel, value=str(self._minAcceptedConfidence))
        self._minAcceptedConfidenceSpinCtrl.SetRange(20, 100)
        self._minAcceptedConfidenceSpinCtrl.Bind(wx.EVT_SPINCTRL, self.OnNumberValueChanged)

        self._minAgreementPercentageLabel = wx.StaticText(self.panel, wx.ID_ANY, 'Minimum agreement percentage (%):', style=wx.ALIGN_RIGHT)
        self._minAgreementPercentageSpinCtrl = wx.SpinCtrl(self.panel, value=str(self._minAgreementPercentage))
        self._minAgreementPercentageSpinCtrl.SetRange(10, 80)
        self._minAgreementPercentageSpinCtrl.Bind(wx.EVT_SPINCTRL, self.OnNumberValueChanged)
        
        self._highResolutionLabel = wx.StaticText ( self.panel, wx.ID_ANY, 'Display high-resolution results:', style=wx.ALIGN_RIGHT )
        self._highResolutionCheckBox = wx.CheckBox ( self.panel, label="", style=wx.ALIGN_LEFT )
        self._highResolutionCheckBox.SetValue ( self._highResolution )
        self._highResolutionCheckBox.Bind ( wx.EVT_CHECKBOX, self.OnHighResolutionChanged )
        
        panelFSizer.AddMany([
            (self._altitudeScaleLabel, 1, wx.EXPAND), (self._altitudeScaleSpinCtrlHBoxSizer, 1, wx.EXPAND),
            (self.ANNFolderButton, 1, wx.EXPAND), (self.ANNFolderLabel, 1, wx.EXPAND),
            (self._filterWindowLabel, 1, wx.EXPAND), (self._filterWindowSpinCtrl, 1, wx.EXPAND|wx.RIGHT, 85),
            (self._minLayerDepthLabel, 1, wx.EXPAND), (self._minLayerDepthSpinCtrl, 1, wx.EXPAND|wx.RIGHT, 85),
            (self._finesseLabel, 1, wx.EXPAND), (self._finesseSpinCtrl, 1, wx.EXPAND|wx.RIGHT, 85),
            (self._minAcceptedConfidenceLabel, 1, wx.EXPAND), (self._minAcceptedConfidenceSpinCtrl, 1, wx.EXPAND|wx.RIGHT, 85),
            (self._minAgreementPercentageLabel, 1, wx.EXPAND), (self._minAgreementPercentageSpinCtrl, 1, wx.EXPAND|wx.RIGHT, 85),
            (self._highResolutionLabel, 1, wx.EXPAND), (self._highResolutionCheckBox, 1, wx.EXPAND)
        ])
        
        self.panel.SetSizer(panelFSizer)            
        
        # Build the bottom buttons:
        
        BottomHBox=wx.BoxSizer(wx.HORIZONTAL)
        
        self.applyBtn = wx.Button(self, label="OK")
        self.applyBtn.Bind(wx.EVT_BUTTON, self.OnApplyBtnClicked)
        self.cancelBtn = wx.Button(self, label="Reset to defaults")
        self.cancelBtn.Bind(wx.EVT_BUTTON, self.OnCancelBtnClicked)
        
        BottomHBox.Add(self.applyBtn, proportion=0, flag=wx.RIGHT, border=5)
        BottomHBox.Add(self.cancelBtn, proportion=0, flag=wx.LEFT, border=5)
        
        VBox=wx.BoxSizer(wx.VERTICAL)
        VBox.Add(self.panel, proportion=1, flag=wx.EXPAND|wx.ALL, border=25)
        VBox.Add(BottomHBox, flag=wx.CENTER|wx.ALL, border=10)
        
        self.SetSizer(VBox)
        
    def OnApplyBtnClicked(self, e):
        '''
        "Apply" button event handler.
        
        Keyword arguments:
        e -- the event.
        '''        
        self.EndModal(wx.ID_OK)
        
        
    def OnCancelBtnClicked(self, e):
        '''
        "Cancel" button event handler.
        
        Keyword arguments:
        e -- the event.
        '''
        self._ANN_Folder = NtGlobals.ANN_FOLDER
        self._filterWindow = NtGlobals.FILTER_WINDOW
        self._minLayerDepth = NtGlobals.MIN_LAYER_DEPTH
        self._finesse = NtGlobals.FINESSE
        self._minAcceptedConfidence = NtGlobals.MIN_ACCEPTED_CONFIDENCE
        self._minAgreementPercentage = NtGlobals.MIN_AGREEMENT_PERCENTAGE
        
        # TODO: Reset the widgets as well!
        
    def OnANNFolderBtnClicked(self, e):
        '''
        "Select folder" toolbar tool event handler.
        
        Keyword arguments:
        e -- the event.
        '''
        dirDialog = wx.DirDialog(None, "Choose ANN folder:", style=wx.DD_DEFAULT_STYLE)
        
        if dirDialog.ShowModal() == wx.ID_OK:
            # Selected folder changed!
            self.OnANNFolderChanged(dirDialog.GetPath())
            
        dirDialog.Destroy()
            
    def OnANNFolderChanged(self, newfolder):
        self._ANN_Folder = newfolder
        self.ANNFolderLabel.SetLabel(newfolder)
        
    def OnNumberValueChanged(self, e):
        obj = e.GetEventObject()
        if obj is self._filterWindowSpinCtrl:
            self._filterWindow = obj.GetValue()
        elif obj is self._minLayerDepthSpinCtrl:
            self._minLayerDepth = obj.GetValue()
        elif obj is self._finesseSpinCtrl:
            self._finesse = obj.GetValue()
        elif obj is self._minAcceptedConfidenceSpinCtrl:
            self._minAcceptedConfidence = obj.GetValue()
        elif obj is self._minAgreementPercentageSpinCtrl:
            self._minAgreementPercentage = obj.GetValue()
        elif obj is self._altitudeScaleLowSpinCtrl:
            self._altitudeScaleLow = obj.GetValue()
            self._altitudeScaleHighSpinCtrl.SetRange(self._altitudeScaleLow+1, 15)
        elif obj is self._altitudeScaleHighSpinCtrl:
            self._altitudeScaleHigh = obj.GetValue()
            self._altitudeScaleLowSpinCtrl.SetRange(0, self._altitudeScaleHigh-1)
            
    def OnHighResolutionChanged( self, e ):
        self._highResolution = e.GetEventObject().GetValue()
            
    def GetSettings(self):
        return {
            'ANN_Folder': self._ANN_Folder,
            'filterWindow': self._filterWindow,
            'minLayerDepth': self._minLayerDepth,
            'finesse': self._finesse,
            'minAcceptedConfidence': self._minAcceptedConfidence / 100.0,
            'minAgreementPercentage': self._minAgreementPercentage / 100.0,
            'altitudeScaleLow': self._altitudeScaleLow,
            'altitudeScaleHigh': self._altitudeScaleHigh,
            'highResolution': self._highResolution
        }
        
class NtFigure(FigureCanvas):
    '''
    Construct plot figures wx widgets from plot data.
    '''
    def __init__(self, parent, id=-1, data=[]):
        '''
        Use to construct a new NtFigure object.
        
        Keyword arguments:
        parent -- wx.Window object this NtFigure object will belong to.
        id -- wxPython ID for this NtFigure object (optional).
        data -- list of tuples to generate a plot (optional).
        '''        
        self.figure = Figure()
        # self.axes = self.figure.add_axes([0, 0, 1, 1])
        self.axes = self.figure.add_subplot(111)
        
        for hdata, vdata in data:
            self.axes.plot(hdata, vdata)
        
        super(NtFigure, self).__init__(parent, id, self.figure)
        
        self.SetMinSize(wx.Size(100, 100))
        self.figure.set_facecolor('white')
        
        self._parent = parent
        self._hasAerosolLegend = False
        
        self._hatching = None
        
        self.mpl_connect('resize_event', self.OnResize)
        
    def TightLayout(self):
        try:
            if self._hasAerosolLegend:
                padRight = 2/self.figure.get_size_inches()[0]
                self.figure.tight_layout(rect=(0, 0, 1-padRight ,1))
            else:
                self.figure.tight_layout()
        except ValueError:
            pass
        
    def OnResize(self, event):
        self.TightLayout()
        
    def Clear(self, settings={}):
        '''
        Clear the figure.
        '''
        self.figure.clf()
        self.axes = self.figure.add_subplot(111)
        
        for key, value in settings.items():
            if key == 'xaxis':
                if value == None:
                    self.axes.get_xaxis().set_visible(False)
                elif isinstance(value, dict):
                
                    self.axes.tick_params(axis='x', labelsize='x-small')
                    self.axes.xaxis.label.set_fontsize('small')
                    
                    for xkey, xvalue in value.items():
                        if xkey == 'limits':
                            self.axes.set_xlim(xvalue)
                        elif xkey == 'label':
                            self.axes.set_xlabel(xvalue)
                        elif xkey == 'color':
                            self.axes.get_xaxis().label.set_color(xvalue)
                            self.axes.spines['bottom'].set_color(xvalue)
                            self.axes.tick_params(axis='x', colors=xvalue)
                        elif xkey == 'format':
                            if xvalue == 'scientific':
                                formatter = FuncFormatter(FormatScientific)
                                self.axes.xaxis.set_major_formatter(formatter)
                elif isinstance(value, list):
                    for xaxis in value[1:]:
                        ax = self.axes.twiny()
                        
                        ax.tick_params(axis='x', labelsize='x-small')
                        ax.xaxis.label.set_fontsize('small')
                        
                        for xkey, xvalue in xaxis.items():                             
                            if xkey == 'limits':
                                ax.set_xlim(xvalue)
                            elif xkey == 'label':
                                ax.set_xlabel(xvalue)
                            elif xkey == 'color':
                                ax.get_xaxis().label.set_color(xvalue)
                                ax.spines['top'].set_color(xvalue)
                                ax.tick_params(axis='x', colors=xvalue)
                            elif xkey == 'format':
                                if xvalue == 'scientific':
                                    formatter = FuncFormatter(FormatScientific)
                                    ax.xaxis.set_major_formatter(formatter)
                                
                    self.axes.tick_params(axis='x', labelsize='x-small')
                    self.axes.xaxis.label.set_fontsize('small')
                    
                    for xkey, xvalue in value[0].items():
                        if xkey == 'limits':
                            self.axes.set_xlim(xvalue)
                        elif xkey == 'label':
                            self.axes.set_xlabel(xvalue)
                        elif xkey == 'color':
                            self.axes.get_xaxis().label.set_color(xvalue)
                            self.axes.spines['bottom'].set_color(xvalue)
                            self.axes.tick_params(axis='x', colors=xvalue)
                        elif xkey == 'format':
                            if xvalue == 'scientific':
                                formatter = FuncFormatter(FormatScientific)
                                self.axes.xaxis.set_major_formatter(formatter)
            elif key == 'yaxis':
                if value == None:
                    self.axes.get_yaxis().set_visible(False)
                elif isinstance(value, dict):
                    
                    self.axes.tick_params(axis='y', labelsize='x-small')
                    
                    for ykey, yvalue in value.items():
                        if ykey == 'limits':
                            self.axes.set_ylim(yvalue)
                        elif ykey == 'label':
                            self.axes.set_ylabel(yvalue)
                        elif ykey == 'color':
                            self.axes.get_yaxis().label.set_color(yvalue)
                            self.axes.spines['left'].set_color(yvalue)
                            self.axes.tick_params(axis='y', colors=yvalue)
                elif isinstance(value, list):
                    for yaxis in value[1:]:
                        ay = self.axes.twinx()
                        
                        ay.tick_params(axis='y', labelsize='x-small')
                        
                        for ykey, yvalue in yaxis.items(): 
                            if ykey == 'limits':
                                ay.set_ylim(yvalue)
                            elif ykey == 'label':
                                ay.set_ylabel(yvalue)
                            elif ykey == 'color':
                                ay.get_yaxis().label.set_color(yvalue)
                                ay.spines['right'].set_color(yvalue)
                                ay.tick_params(axis='y', colors=yvalue)
                                
                    self.axes.tick_params(axis='y', labelsize='x-small')
                    
                    for ykey, yvalue in value[0].items():
                        if ykey == 'limits':
                            self.axes.set_ylim(yvalue)
                        elif ykey == 'label':
                            self.axes.set_ylabel(yvalue)
                        elif ykey == 'color':
                            self.axes.get_yaxis().label.set_color(yvalue)
                            self.axes.spines['left'].set_color(yvalue)
                            self.axes.tick_params(axis='y', colors=yvalue)
                            
        if 'aerosolLegend' in settings.keys():
            if settings['aerosolLegend']:
                
                self._hasAerosolLegend = True
                
                # Add legend
                legend_patches = []
                legend_labels = []
                for aerosol, color in NtGlobals.AEROSOL_COLORS.items():
                    legend_patches.append(mpatches.Patch(color=color, label=aerosol))
                    legend_labels.append(aerosol)
                    
                self.figure.legend(
                    handles=legend_patches,
                    labels=legend_labels,
                    loc='center right',
                    fontsize='x-small',
                    ncol=1,
                    frameon=False)
                    
                self.TightLayout()
        
        self._parent.Layout()
        
    def Plot(self, data):
        '''
        Plot data inside the figure.
        
        Keyword arguments:
        data -- list of tuples as (xdata, ydata).
        '''
        for xdata, ydata in data:
            self.axes.plot(xdata, ydata)
            
        self._parent.Layout()
        
    def PlotAerosols(self, **kwargs):
        acceptedArgs = ['ydata', 'ylabel', 'ylimits', 'aerosols', 'clear']
        
        ydata = None
        ylabel = ''
        ylimits = None
        xlimits = None
        axes=[]
        clear = False
        
        for arg, value in kwargs.iteritems():
            # Parse arguments and raise TypeError if any
            # unknown parameters are found.
            if arg not in acceptedArgs:
                raise TypeError('Invalid argument in NtFigure.PlotDoubleHorizontal(): ' + arg)
                
            # Read known parameters and set member variables.
            if arg == 'ydata':
                ydata = value
            elif arg == 'ylabel':
                ylabel = value
            elif arg == 'aerosols':
                aerosols = value
            elif arg == 'clear':
                clear = value
            elif arg == 'ylimits':
                ylimits = value
                
        if len(aerosols) < 1:
            # Nothing to do!
            return
                
        if ydata is None:
            return
            
        if clear == True:
            self.figure.clf()
            self.axes = self.figure.add_subplot(111)
        
        defaultColors = ['Red', 'Green', 'Blue', 'Yellow', 'Orange', 'Brown', 'Pink', 'Black']
        
        legend_handles=[]
        
        if ylimits != None:
            self.axes.set_ylim(ylimits)
        else:
            self.axes.set_ylim((ydata[0], ydata[-1]))
        
        i = 0
        for mdata in aerosols:
            name = mdata['name']
            
            color = NtGlobals.AEROSOL_COLORS[name]
                
            if color == '#FFFFFF':
                # Skip drawing 'empty' boxes.
                continue
            
            linestyle='solid'
            if 'linestyle' in mdata.keys():
                linestyle=mdata['linestyle']
                
            yd = ydata
            if 'ydata' in mdata.keys():
                yd = mdata['ydata']
            
            handle, = self.axes.plot([0, 0], [mdata['bottom'], mdata['top']], color=color, label=name, linestyle=linestyle)
            
            self.axes.axhspan(mdata['bottom'], mdata['top'], facecolor=color, edgecolor='none')
            
            i += 1
            
        loc = 1
        
        leg = self.axes.legend(loc=loc, fontsize='medium')
            
        self.axes.set_ylabel(ylabel)
        self.axes.get_xaxis().set_visible(False)
        
        self.axes.tick_params(axis='x', labelsize='x-small')
        self.axes.tick_params(axis='y', labelsize='x-small')
        
        self.TightLayout()
        self._parent.Layout()
            
    def PlotMultipleHorizontal(self, **kwargs):
        acceptedArgs = ['ydata', 'ylabel', 'ylimits', 'xlimits', 'axes', 'clear']
        
        ydata = None
        ylabel = ''
        ylimits = None
        xlimits = None
        axes=[]
        clear = False
        
        for arg, value in kwargs.iteritems():
            # Parse arguments and raise TypeError if any
            # unknown parameters are found.
            if arg not in acceptedArgs:
                raise TypeError('Invalid argument in NtFigure.PlotDoubleHorizontal(): ' + arg)
                
            # Read known parameters and set member variables.
            if arg == 'ydata':
                ydata = value
            elif arg == 'ylabel':
                ylabel = value
            elif arg == 'ylimits':
                ylimits = value
            elif arg == 'xlimits':
                xlimits = value
            elif arg == 'axes':
                axes = value
            elif arg == 'clear':
                clear = value
                
        if len(axes) < 1:
            # Nothing to do!
            return
                
        if ydata is None:
            return
            
        if clear == True:
            self.figure.clf()
            self.axes = self.figure.add_subplot(111)
        
        defaultColors = ['Red', 'Green', 'Blue', 'Yellow', 'Orange', 'Brown', 'Pink', 'Black']
        
        # TODO: Implement setting xlimit and ylimit!
        
        legend_handles=[]
        
        if ylimits != None:
            self.axes.set_ylim(ylimits)
        else:
            self.axes.set_ylim(ydata[0], ydata[-1])
        
        i = 0
        for mdata in axes[0]['plots']:
            name = mdata['name']
            
            color = defaultColors[i % len(defaultColors)]
            if 'color' in mdata.keys():
                color = mdata['color']
            
            linestyle='solid'
            if 'linestyle' in mdata.keys():
                linestyle=mdata['linestyle']
                
            yd = ydata
            if 'ydata' in mdata.keys():
                yd = mdata['ydata']
            
            handle, = self.axes.plot(mdata['xdata'], yd, color=color, label=name, linestyle=linestyle)
            
            if 'xerror' in mdata.keys():
                self.axes.fill_betweenx(yd, [x-e for x, e in zip(mdata['xdata'], mdata['xerror'])], [x+e for x, e in zip(mdata['xdata'], mdata['xerror'])], alpha=0.2, color=color, linestyle=linestyle)
            
            i += 1
            
        self.axes.xaxis.label.set_fontsize('small')
            
        loc = 4
        if 'loc' in axes[0].keys():
            loc = axes[0]['loc']
            
        if len(axes[0]['plots']) > 0:
            leg = self.axes.legend(loc=loc, fontsize='x-small')
        
        if 'color' in axes[0].keys():
            self.axes.spines['bottom'].set_color(axes[0]['color'])
            self.axes.tick_params(axis='x', colors=axes[0]['color'])
            self.axes.xaxis.label.set_color(axes[0]['color'])
            
            [text.set_color(axes[0]['color']) for text in leg.get_texts()]
            
            leg.get_frame().set_edgecolor(axes[0]['color'])
        
        if 'xlimits' in axes[0].keys():
            self.axes.set_xlim(axes[0]['xlimits'])
            
        if 'label' in axes[0].keys():
            self.axes.set_xlabel(axes[0]['label'])
            
        if 'format' in axes[0].keys():
            if axes[0]['format'] == 'scientific':
                formatter = FuncFormatter(FormatScientific)
                self.axes.xaxis.set_major_formatter(formatter)
            
        self.axes.set_ylabel(ylabel)
        
        self.axes.tick_params(axis='x', labelsize='x-small')
        self.axes.tick_params(axis='y', labelsize='x-small')
        
        self.axes.grid(True)
        
        if len(axes) < 2:
            self.figure.subplots_adjust(top=0.875)
        
        for ax in axes[1:]:
            i = 0
            
            axis=self.axes.twiny()
            
            for mdata in ax['plots']:
                name = mdata['name']
                color = defaultColors[i % len(defaultColors)]
                if 'color' in mdata.keys():
                    color = mdata['color']

                axis.grid(False)
                
                linestyle='solid'
                if 'linestyle' in mdata.keys():
                    linestyle=mdata['linestyle']
                    
                yd = ydata
                if 'ydata' in mdata.keys():
                    yd = mdata['ydata']
                
                handle, = axis.plot(mdata['xdata'], yd, color=color, label=name, linestyle=linestyle)
                
                if 'xerror' in mdata.keys():
                    axis.fill_betweenx(yd, [x-e for x, e in zip(mdata['xdata'], mdata['xerror'])], [x+e for x, e in zip(mdata['xdata'], mdata['xerror'])], alpha=0.2, color=color, linestyle=linestyle)
                
                i += 1
                
            axis.xaxis.label.set_fontsize('small')
                
            loc = 3
            if 'loc' in ax.keys():
                loc = ax['loc']
                
            if len(ax['plots']) > 0:
                leg = axis.legend(loc=loc, fontsize='x-small')
            
            if 'color' in ax.keys():
                axis.spines['top'].set_color(ax['color'])
                axis.tick_params(axis='x', colors=ax['color'])
                axis.xaxis.label.set_color(ax['color'])
                
                [text.set_color(ax['color']) for text in leg.get_texts()]
                    
                leg.get_frame().set_edgecolor(ax['color'])
            
            if 'xlimits' in ax.keys():
                axis.set_xlim(ax['xlimits'])
                
            if 'label' in ax.keys():
                axis.set_xlabel(ax['label'])
                
            if 'format' in ax.keys():
                if ax['format'] == 'scientific':
                    formatter = FuncFormatter(FormatScientific)
                    axis.xaxis.set_major_formatter(formatter)
                    
            if ylimits != None:
                axis.set_ylim(ylimits)
            else:
                axis.set_ylim(ydata[0], ydata[-1])
                    
            axis.tick_params(axis='x', labelsize='x-small')
         
        # self.TightLayout()            
        self._parent.Layout()
        
    def PlotAerosolTimeseries(self, datasets, ylimits=None, activeDataset='', highResolution = True):
        self.figure.clf()
        self.axes = self.figure.add_subplot(111)
        
        datasetsCount = len(datasets)
        
        hmax = 1
                    
        if ylimits != None:
            self.axes.set_ylim(ylimits)
        else:
            for data in datasets.values():
                if 'results' not in data.keys():
                    # Don't plot datasets we don't
                    # have results for.
                    continue
                    
                for layer in data['results']:
                    if layer['top'] > hmax:
                        hmax = layer['top']
                        
            self.axes.set_ylim(0, hmax / 1000.0)
            
        self.axes.set_xlim(0, datasetsCount)
        
        # Start from -1 to increment value before using it.
        # Use float for setting relative horizontal boundaries
        # inside the figure (0 - left, 1 - right).
        datasetNum = -1.0
        
        xlabels = []
        
        self._hatching = None
        
        for dataset, data in datasets.items():
            # Increment datasetNum before checking for a
            # valid result to properly align bars on figure.
            datasetNum += 1.0
            
            xlabels.append(dataset)
            # xlabels.append('') # Insert a blank label between datasets
            
            if dataset == activeDataset:
                self._hatching = self.axes.fill_between([2*datasetNum, 2*(datasetNum+1)], [self.axes.get_ylim()[1]]*2, color="none", edgecolor='black', linewidth=2.4)
            
            if 'results' not in data.keys():
                # Don't plot datasets we don't
                # have results for.
                continue
                
            for layer in data['results']:
                
                aerosol = layer['L_result'][0]
            
                if layer['H_result'][0] != 'Unknown' and layer['H_result'][0] != 'N/A' and highResolution:
                    aerosol = layer['H_result'][0]
                
                # if aerosol == 'N/A':
                #    continue
                
                color = NtGlobals.AEROSOL_COLORS[aerosol]
                
                self.axes.axhspan(ymin=layer['bottom'] / 1000.0, ymax=layer['top'] / 1000.0, xmin=datasetNum/datasetsCount, xmax=(datasetNum+1)/datasetsCount, facecolor=color, edgecolor='none')
                
        # Optimisation for large number of datasets:
        # do not display the entire set of labels.
        if len(datasets) > 15:
            labelPositions = GenerateLabelPositions(4, len(datasets)-1)
            
            labelPositions.insert(0, 0)
            labelPositions.append(len(datasets)-1)
            
            for i in range(0, len(xlabels)):
                if i not in labelPositions:
                    xlabels[i] = ''
        
        self.axes.set_ylabel('Altitude (km)')
        
        self.axes.tick_params(axis='x', which='minor', labelsize='x-small')
        self.axes.tick_params(axis='x', which='major', labelsize=0)
        self.axes.tick_params(axis='y', labelsize='x-small')
        
        self.axes.set_xticks([2*x+1 for x in arange(0, datasetsCount)], minor=True)
        self.axes.set_xticklabels(xlabels, minor=True, rotation=60)
        
        self.axes.set_xticks([2*x for x in arange(0, datasetsCount+1)], minor=False)
        self.axes.set_xticklabels([''] * datasetsCount, minor=False)
        
        if datasetsCount < 50:
            self.axes.grid(True)
        
        self._hasAerosolLegend = True
        
        # Add legend
        legend_patches = []
        legend_labels = []
        
        for aerosol, color in NtGlobals.AEROSOL_COLORS.items():
            legend_patches.append(mpatches.Patch(color=color, label=aerosol))
            legend_labels.append(aerosol)
            
        self.figure.legend(
            handles=legend_patches,
            labels=legend_labels,
            loc='center right',
            fontsize='x-small',
            ncol=1,
            frameon=False)
          
        self.TightLayout()
        self._parent.Layout()
    
    def DisplayText(self, text):
        self.axes.text(.5, .5, text, horizontalalignment='center', verticalalignment='center', transform=self.axes.transAxes, fontsize='large')
        
        self._parent.Layout()
    
    def FakeFill(self):
        t = arange(0.0, 3.0, 0.01)
        s = sin(2 * pi * t)
        
        self.axes.plot(t, s)
        
    def GetHatching(self):
        return self._hatching
        
    def DrawIdle(self):
        self.draw_idle()
   
class NtFigurePopupMenu(wx.Menu):

    def __init__(self, parent):
        super(NtFigurePopupMenu, self).__init__()
        
        self.parent = parent
        
        smi = wx.MenuItem(self, wx.ID_ANY, 'Save as...')
        self.AppendItem(smi)
        self.Bind(wx.EVT_MENU, self.OnSave, smi)
        
    def OnSave(self, e):
        saveFileDialog = wx.FileDialog(self.parent._parent, "Save plot as PNG", "", "",
                                   "PNG files (*.png)|*.png", wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
                                   
        if saveFileDialog.ShowModal() != wx.ID_OK:
            # The user changed his mind or something strange
            # has happened.
            return
            
        if self.parent.GetHatching() != None:
            self.parent.GetHatching().set_alpha(0)
            
        self.parent.figure.savefig(saveFileDialog.GetPath(), dpi=600)
        
        if self.parent.GetHatching() != None:
            self.parent.GetHatching().set_alpha(1)
            self.parent.DrawIdle()
   
def FormatScientific(x, pos):
    return "%.0e" % x
    
def GenerateLabelPositions(maxlevel, end, start=0, level=0):        
    if level == maxlevel:
        return []
        
    result = GenerateLabelPositions(maxlevel, (start+end)/2-1, start, level+1)
    result += [(start+end)/2]
    result += GenerateLabelPositions(maxlevel, end, (start+end)/2+1, level+1)
    
    return result